"""

"""

import logging
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, List, Optional, Type, Union

import numpy as np
import pandas as pd
import pytz

import arthurai.core.util as core_util
from arthurai.common.constants import InferenceType, TimestampInferenceType, ValueType
from arthurai.common.exceptions import (
    UnexpectedValueError,
    UserTypeError,
    UserValueError,
)

logger = logging.getLogger(__name__)


"""
Only client facing functions should live in this module. If a util  is used for a specific package and not client facing
it should be added to that packages util.py file.
"""


def format_timestamps(
    inferences: List[Dict[str, Any]],
    tz: Optional[Union[timezone, str]] = None,
    timestamp_attributes: Optional[List[str]] = None,
    time_series_attributes: Optional[List[str]] = None,
    is_reference_data: bool = False,
) -> List[Dict[str, Any]]:
    """Checks list of json inferences to ensure their timestamps have been converted to ISO8601 conventions.

    :param inferences: inferences from input to `send_inferences()` function or output from `nest_reference_data()`
        function. This enforces requirement that the inference has an `"inference_data"` or `"reference_data"` field.
    :param tz: datetime timezone object or timezone string
    :param timestamp_attributes: List of attributes of the model of timestamp type
    :param time_series_attributes: List of attributes of the model of time series type
    :param is_reference_data: True if data to be formatted is reference data, False if inference data or ground truth
     data

    :return: updated list of json inferences with ISO8601 formatted timestamps
    :raises TypeError: timestamp is not of type `datetime.datetime`
    :raises ValueError: timestamp is not timezone aware and no location data is provided to remedy
    """
    inference_key = (
        InferenceType.INFERENCE_DATA
        if not is_reference_data
        else InferenceType.REFERENCE_DATA
    )
    if len(inferences) > 0:
        for inference in inferences:
            if TimestampInferenceType.INFERENCE_TIMESTAMP in inference:
                inference_timestamp = inference[
                    TimestampInferenceType.INFERENCE_TIMESTAMP
                ]
                inference[TimestampInferenceType.INFERENCE_TIMESTAMP] = (
                    format_timestamp(inference_timestamp, tz)
                )
            if TimestampInferenceType.GROUND_TRUTH_TIMESTAMP in inference:
                gt_timestamp = inference[TimestampInferenceType.GROUND_TRUTH_TIMESTAMP]
                inference[TimestampInferenceType.GROUND_TRUTH_TIMESTAMP] = (
                    format_timestamp(gt_timestamp, tz)
                )
            if timestamp_attributes is not None:
                for attr in timestamp_attributes:
                    attr_timestamp = inference[inference_key].get(attr)
                    if attr_timestamp is not None:
                        inference[inference_key][attr] = format_timestamp(
                            attr_timestamp, tz
                        )
            if time_series_attributes is not None:
                inference[inference_key] = format_time_series_attr_timestamps(
                    inference, time_series_attributes, tz, is_reference_data
                )

    return inferences


def format_timestamp(
    timestamp: Union[datetime, str],
    tz: Optional[Union[timezone, str]] = None,
    col_name: str = "",
) -> str:
    """Check if timestamp is time aware and convert to ISO8601 conventioned string.

    This helper function converts datetime objects into timezone aware ISO8601 strings, which is necessary when sending
    JSON to Arthur's backend. If `timestamp` argument is timezone aware, no `location` needs to be provided; otherwise,
    a string pytz `location` like "US/Eastern" needs to be provided to establish timezone. String `timestamp` are
    supported for backwards compatability and for simplicity are assumed to already be in UTC format, but string support
    will be deprecated.

    :param timestamp: timestamp to format
    :param tz: datetime timezone object or timezone string
    :param col_name: Optionally used to specify a column the timestamp originates from in the error message

    :return: ISO8601 formatted timestamp
    :raises TypeError: timestamp is not of type `datetime.datetime`
    :raises ValueError: timestamp is not timezone aware and no location data is provided to remedy
    """
    if col_name != "":
        col_name = " in column " + col_name

    if isinstance(timestamp, datetime):
        is_timezone_aware = (
            timestamp.tzinfo is not None
            and timestamp.tzinfo.utcoffset(timestamp) is not None
        )
        if is_timezone_aware:
            return timestamp.isoformat()
        else:
            # Convert string tzinfo into a datetime object
            if tz and isinstance(tz, str):
                localization = pytz.timezone(tz)
                return localization.localize(timestamp).astimezone(pytz.utc).isoformat()
            # Convert timezone object into a datetime object
            if tz and isinstance(tz, timezone):
                return timestamp.replace(tzinfo=tz).astimezone(pytz.utc).isoformat()
            else:
                raise UserValueError(
                    f"Timestamps{col_name} should be timezone aware. "
                    f"Please specify a pytz tz or set the timezone in the datetime object"
                )
    elif isinstance(timestamp, str):
        if is_valid_datetime_string(timestamp):
            return timestamp
        else:
            raise UserValueError(
                f"Timestamp '{timestamp}'{col_name} is not in ISO8601 format."
            )
    else:
        raise UserTypeError(
            f"Timestamps{col_name} must be of type datetime or string. Provided type: {str(type(timestamp))}"
        )


def format_time_series_attr_timestamps(
    inference: Dict[str, Any],
    time_series_attributes: List[str],
    tz: Optional[Union[timezone, str]] = None,
    is_reference_data: bool = False,
) -> Dict[str, Any]:
    """Formats time series timestamps in json inference to ISO8601 conventions.

    If a passed time series attribute is missing or the attribute is missing a timestamp field it will be skipped
    without raising an error.

    :param inference: Single inference from input to `send_inferences()` function or output from `nest_reference_data()`
        function. This enforces requirement that the inference has an `"inference_data"` or `"reference_data"` field.
    :param time_series_attributes: List of names of attributes of the model of time series value type
    :param tz: datetime timezone object or timezone string
    :param is_reference_data: True if reformatting reference data, False if reformatting inference data

    :returns: updated json inference_data or reference_data field
    :raises TypeError: timestamp is not of type `datetime.datetime`
    :raises ValueError: timestamp is not timezone aware and no location data is provided to remedy or inference
        does not match expected formatting
    """
    if not isinstance(inference, dict):
        raise UserValueError("inference is not a dictionary")

    inference_key = (
        InferenceType.INFERENCE_DATA
        if not is_reference_data
        else InferenceType.REFERENCE_DATA
    )
    for attr in time_series_attributes:
        if inference.get(inference_key) is None or not isinstance(
            inference.get(inference_key), dict
        ):
            raise UserValueError(
                f"inference does not have the expected {inference_key} field"
            )
        time_series_data = inference[inference_key].get(attr)
        if time_series_data is not None and core_util.is_list_like(time_series_data):
            for i in range(len(time_series_data)):
                time_series_timestamp = time_series_data[i].get("timestamp")
                if time_series_timestamp is not None:
                    inference[inference_key][attr][i]["timestamp"] = format_timestamp(
                        time_series_timestamp, tz, col_name=attr
                    )
    return inference[inference_key]


def is_valid_datetime_string(dt_obj: Any) -> bool:
    """ "
    Determines if an object is a string and in correct ISO8601 format

    :param dt_obj: possible datetime object
    :return: bool
    """
    if isinstance(dt_obj, str):
        try:
            timestamp = datetime.fromisoformat(dt_obj.replace("Z", "+00:00"))
            is_timezone_aware = (
                timestamp.tzinfo is not None
                and timestamp.tzinfo.utcoffset(timestamp) is not None
            )

            return True if is_timezone_aware else False
        # Case where string is not in valid iso-format
        except ValueError:
            return False
    else:
        return False


def normal_random_ints_fixed_sum(
    num_values: int, total_sum: int, relative_std_dev: float = 0.5
) -> np.ndarray:
    """
    Return `num_values` roughly normally-distributed integers summing to `total_sum`. Numbers are first sampled from
    the normal distribution, then adjusted to fit the sum. Adjustments are made first by shifting the resulting
    distribution to be at least `total_sum`, then decrementing a subset of the values to match `total_sum`.
    :param num_values: number of integers to return
    :param total_sum: total sum the integers returned
    :param relative_std_dev: the relative standard deviation (std_dev / mean) of the initial distribution.
    :return:
    """
    # first pull from the normal distribution to get some initial values
    mean_value = total_sum / num_values
    std_dev = mean_value * relative_std_dev
    initial_values = np.round(
        np.random.normal(loc=mean_value, scale=std_dev, size=num_values)
    ).astype(np.int64)
    # bump any negatives values up to zero
    positive_values = np.clip(initial_values, a_min=0, a_max=None)

    # shift the distribution to get close to the correct sum
    mean_diff_floor = np.floor((positive_values.sum() - total_sum) / num_values).astype(
        np.int64
    )
    values = positive_values - mean_diff_floor

    # now our difference must be less than the length of the array, and positive
    final_difference = values.sum() - total_sum
    # we can simply subtract 1 from however many indices we need to to reach the exact sum
    inds_to_decrement = np.random.choice(
        num_values, size=final_difference, replace=False
    )
    values[inds_to_decrement] -= 1

    # failsafe check
    if values.sum() != total_sum:
        raise UnexpectedValueError(
            "Computed random integers do not sum to expected total. Please report to Arthur with "
            f"the following debug info: total_sum={total_sum}, num_values={num_values}, "
            f"mean_diff_floor={mean_diff_floor}, final_difference={final_difference}, "
            f"positive_values={positive_values}, inds_to_decrement={inds_to_decrement}"
        )

    return values


def _parse_duration(time_ref: str) -> timedelta:
    """
    Parses a time duration string. Supports days and hours, e.g. "14d", "-3h", etc.
    :param time_ref: the time string to parse
    :return: timedelta of the parsed value
    """
    raw_value, unit = time_ref[:-1], time_ref[-1:]
    try:
        value = int(raw_value)
    except ValueError as e:
        raise UserValueError(
            f"Cannot parse time reference value {raw_value}, must be an integer"
        ) from e
    if unit == "d":
        return timedelta(days=value)
    elif unit == "h":
        return timedelta(hours=value)
    else:
        raise UserValueError(
            f"Cannot parse time reference unit {unit}, only days ('d') and hours ('h') are supported"
        )


# for mocking datetime.now without clobbering the entire datetime class
def _datetime_now(*args, **kwargs):
    return datetime.now(*args, **kwargs)


def generate_timestamps(
    total: int,
    duration: Union[str, timedelta] = "7d",
    end: Union[str, datetime] = "now",
    freq: Union[str, pd.DateOffset] = "D",
    noisiness: float = 0.3,
) -> List[datetime]:
    """
    Generates timestamps over a period of time, defaulting to daily timestamps over the last week. Creates `uniques`
    unique timestamp values evenly spaced between `start` and `end` inclusively. Repeats values if
    total ≠ uniques. This is useful for generating timestamps for test inferences outside of a production setting.

    :param total: the total number of timestamps to generate
    :param end: the final timestamp value
    :param duration: the difference between the first and last timestamp values. day or hour strings (e.g. "7d", "4h")
        and timedelta objects are supported
    :param freq: the frequency with which to generate unique values. See
        `Pandas Offset Aliases <https://pandas.pydata.org/docs/user_guide/timeseries.html#offset-aliases>`_ for values
    :param noisiness: how much to noise to add to the number of timestamps repeated at each interval. specifically, the
        coefficient of variation (standard deviation / mean) to use when selecting the number of times to repeat each
        timestamp
    :raises UserValueError: if an input is invalid or values contain future timestamps and allow_future is False
    :return: a list of generated timestamps
    """
    if end == "now":
        end_dt = _datetime_now(pytz.utc)
    elif not isinstance(end, datetime):
        raise UserTypeError(
            f"'end' should be string 'now' or a datetime object, got type {type(end)} with value {end}"
        )
    else:
        end_dt = end

    if isinstance(duration, str):
        duration = _parse_duration(duration)

    try:
        unique_values = pd.date_range(
            start=(end_dt - duration), end=end_dt, freq=freq
        ).to_pydatetime()
    except ValueError as e:
        raise UserValueError(e) from e
    unique_count = len(unique_values)

    if unique_count > total:
        raise UserValueError(
            f"Timestamps with duration {duration} and frequency {freq} yield {unique_count} unique "
            f"values, but you asked for {total} total values. Total values must be greater than or "
            "equal to the number of unique timestamps."
        )

    # determine how many times each timestamp should be repeated
    repeat_counts = normal_random_ints_fixed_sum(
        unique_count, total, relative_std_dev=noisiness
    )
    timestamps = unique_values.repeat(repeat_counts).tolist()

    return timestamps


def value_type_to_python_type(value_type: ValueType) -> Type:
    if value_type == ValueType.Integer:
        return int
    elif value_type == ValueType.Float:
        return float
    elif value_type == ValueType.Boolean:
        return bool
    elif value_type in (
        ValueType.String,
        ValueType.Unstructured_Text,
        ValueType.Image,
    ):  # image attributes contain strings referencing images
        return str
    elif value_type == ValueType.BoundingBox:
        return list
    else:
        raise UserValueError(f"Can't convert value type {value_type} to Python type")



import json
from dataclasses import asdict
from typing import Union

from dataclasses_json import DataClassJsonMixin


class ArthurBaseJsonDataclass(DataClassJsonMixin):
    @staticmethod
    def clean_nones(d):
        """Helper function to filter out None objects from a json or dictionary representation of an object

        :param d: a Dictionary or Json representation of an ArthurBaseJsonDataclass object
        :return: Dictionary of the object with all None components removed
        """
        if not isinstance(d, (dict, list)):
            return d
        if isinstance(d, list):
            return list(
                filter(
                    lambda item: item is not None,
                    map(ArthurBaseJsonDataclass.clean_nones, d),
                )
            )
        return {
            k: v
            for k, v in (
                (k, ArthurBaseJsonDataclass.clean_nones(v)) for k, v in d.items()
            )
            if v is not None
        }

    def to_json(self, skip_none=True) -> str:  # type: ignore
        """Creates a json representation of this object

        This function can be applied to any extension of the ArthurBaseJsonDataClass

        :return: json of object data
        """
        obj_dict = asdict(self)
        if skip_none:
            obj_dict = self.clean_nones(obj_dict)
        return json.dumps(obj_dict)

    def to_dict(self, skip_none=True) -> dict:
        """Creates a dictionary representation of this object

        This function can be applied to any extension of the ArthurBaseJsonDataClass

        :return: Dictionary of object data
        """
        obj_dict = asdict(self)
        if skip_none:
            obj_dict = self.clean_nones(obj_dict)
        return obj_dict


NumberType = Union[int, float]



import glob
import importlib
import logging
import os
import shutil
import sys
import tempfile
from pathlib import PurePath
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Tuple
from zipfile import ZipFile

# need dill to be able to pickle LIME
import dill
import docker
import pandas as pd
import pkg_resources

from arthurai.common.constants import InputType, Stage, ValueType
from arthurai.common.exceptions import (
    ArthurUserError,
    ExpectedParameterNotFoundError,
    UnexpectedTypeError,
    UserValueError,
)
from arthurai.core.util import NumpyEncoder
from arthurai.explainability.arthur_explainer import ArthurExplainer

#  imports ArthurModel for type checking, required due to circular import
from arthurai.explainability.validation import (
    validate_predicted_attribute_order_matches_remote,
)

if TYPE_CHECKING:
    from arthurai.core.models import ArthurModel

logger = logging.getLogger(__name__)


class ExplanationPackager(object):
    """Manages the project source code and explainer for this model.
    Intended only for internal use. Used to create explainer object, package files and format config for the
    update enrichments API endpoint.

    See model.update_enrichments() for more details on usage
    """

    # fields required to generate explainability files
    FILE_FIELDS = [
        "project_directory",
        "user_predict_function_import_path",
        "requirements_file",
    ]

    def __init__(self, model: "ArthurModel", **kwargs):
        """Constructor for ExplanationPackager. Explanation packager needs to handle case of
        user setting files, and user just updating config. Because of this, don't raise any
        exceptions in constructor for missing required fields. Instead check and raise the exception in
        the function that needs the field.
        """
        self._model = model

        # use all kwargs so its easy to call constructor using config dict in model.update_enrichments
        self.df: pd.DataFrame = kwargs.get("df")
        self.project_directory: Optional[str] = kwargs.get("project_directory")
        self.streaming_explainability_enabled = kwargs.get(
            "streaming_explainability_enabled"
        )
        self.ignore_dirs = set(kwargs.get("ignore_dirs", []))
        self.user_predict_function_import_path: Optional[str] = kwargs.get(
            "user_predict_function_import_path"
        )
        self.requirements_file = kwargs.get("requirements_file")
        self.python_version = kwargs.get("python_version")
        self.sdk_version = kwargs.get("sdk_version")
        self.explanation_nsamples = kwargs.get("explanation_nsamples")
        self.explanation_algo = kwargs.get("explanation_algo")
        self.model_server_num_cpu = kwargs.get("model_server_num_cpu")
        self.model_server_memory = kwargs.get("model_server_memory")
        self.model_server_max_replicas = kwargs.get("model_server_max_replicas")
        self.inference_consumer_num_cpu = kwargs.get("inference_consumer_num_cpu")
        self.inference_consumer_memory = kwargs.get("inference_consumer_memory")
        self.inference_consumer_thread_pool_size = kwargs.get(
            "inference_consumer_thread_pool_size"
        )
        self.inference_consumer_score_percent = kwargs.get(
            "inference_consumer_score_percent"
        )
        self.text_delimiter = model.text_delimiter
        self.shap_nrows = kwargs.get("shap_nrows", 20)
        self.seed = kwargs.get("seed", None)

        # Get all hidden files and add them to the ignore_dirs list
        if self.project_directory:
            hidden_files_temp = glob.glob(os.path.join(self.project_directory, ".*"))
            hidden_files = set([path.split("/")[-1] for path in hidden_files_temp])
            self.ignore_dirs = self.ignore_dirs.union(hidden_files)

        # explanation_algo is saved on the model object, and determines which explainer to use
        # however we always create both lime and shap, if possible, regardless of explanation_algo
        # this is to more easily allow switching between the 2
        self.enable_shap = self._model.input_type != InputType.NLP
        self.enable_lime = True
        if (
            self._model.input_type == InputType.NLP
            or self._model.input_type == InputType.Image
        ) and self.explanation_algo != ArthurExplainer.LIME:
            logger.warning(
                f"Only lime is supported for {self._model.input_type} explainabilty but shap was passed. "
                "Automatically setting explanation_algo to lime"
            )
            self.explanation_algo = ArthurExplainer.LIME

        self.explainer: ArthurExplainer = None
        self.zipfile: Optional[PurePath] = None

        self.dirname = "model_files"
        self.volume = os.path.join(tempfile.TemporaryDirectory().name, self.dirname)
        self.bind_path = "/mnt/model_files"
        self.test_script = "test.sh"
        self.example: Optional[List] = None

    def contains_file_fields(self) -> Tuple[bool, bool]:
        """Returns (onePresent, allPresent)
        onePresent is true if at least one file field was supplied
        AllPresent is true if all file fields are supplied
        """
        # NOTE if adding new file field check here, ensure its added to FILE_FIELDS at top of file
        checks = []
        checks.append(self.project_directory is not None)
        checks.append(self.user_predict_function_import_path is not None)
        checks.append(self.requirements_file is not None)

        return (any(checks), all(checks))

    def build_categorical_lookup_table(self) -> List[Optional[Dict[str, int]]]:
        """Builds the lookup table that ArthurExplainer needs for encoding strings"""
        num_attributes = len(self._model.get_attributes(Stage.ModelPipelineInput))
        label_mapping: List[Optional[Dict[str, int]]] = [
            None for _ in range(num_attributes)
        ]
        sorted_attr = sorted(
            self._model.get_attributes(Stage.ModelPipelineInput),
            key=lambda x: x.position,
        )
        for i, attr in enumerate(sorted_attr):
            if attr.categorical and attr.value_type == ValueType.String:
                # When categories are available as strings, convert them to numericals for training the explainer.
                if attr.categories is None or len(attr.categories) == 0:
                    # This is a hack. In the case that there are many categories,
                    # we don't want to save the categories in the backend but we need the categories to train the explainer.
                    # Therefore, we access the categories from the sample dataframe.
                    label_mapping[i] = {
                        val: j for j, val in enumerate(self.df.iloc[:, i].unique())
                    }
                else:
                    label_mapping[i] = {
                        val.value: j for j, val in enumerate(attr.categories)
                    }
        return label_mapping

    def create(self) -> None:
        """Create an explainer for this model.

        :param: df: the training data for the explainer
        :param: shap_nrows: number of rows to sample the dataframe passed in
        """
        # this should never happen since we check for this at model onboarding
        if self._model.input_type == InputType.NLP and (
            self.text_delimiter is None or self.text_delimiter == ""
        ):
            raise ExpectedParameterNotFoundError(
                f"Missing text delimiter. {InputType.NLP} require setting a text delimiter."
            )

        pipeline_attributes = self._model.get_attributes(Stage.ModelPipelineInput)
        pipeline_attributes_in_order = sorted(
            pipeline_attributes, key=lambda x: x.position
        )
        predicted_attrs = self._model.get_attributes(Stage.PredictedValue)

        # get categorical attr indices
        categorical_attr_indices = [
            i for i, attr in enumerate(pipeline_attributes_in_order) if attr.categorical
        ]

        # pull out pipeline features from data
        # dataframe is not required for image models, only tabular and nlp
        if self.df is not None:
            attr_names = [attr.name for attr in pipeline_attributes_in_order]
            explanation_training_data = self.df[attr_names]
            self.example = list(explanation_training_data.iloc[0, :])

            pred_cols_in_df = [
                pred_attr.name
                for pred_attr in predicted_attrs
                if pred_attr.name in self.df.columns
            ]
            predictions_df = self.df[pred_cols_in_df]
        else:
            explanation_training_data = None
            predictions_df = None

        if self.project_directory is None:
            raise UnexpectedTypeError("project_directory is None")
        if not os.path.isabs(self.project_directory):
            raise UserValueError(
                f"Failed to find '{self.project_directory}'. Please ensure this is an absolute path."
            )
        sys.path.insert(0, self.project_directory)
        current_dir = os.getcwd()
        os.chdir(self.project_directory)
        try:
            req_file_errors = self.validate_requirements_add_sklearn()
            if len(req_file_errors) > 0:
                error_str = "\n" + "\n".join(req_file_errors)
                raise UserValueError(
                    f"Failed to validate requirements file: {error_str}"
                )

            if self.user_predict_function_import_path is None:
                raise UnexpectedTypeError("user_predict_function_import_path is None")
            try:
                customer_model: Any = importlib.import_module(
                    self.user_predict_function_import_path
                )
            except ModuleNotFoundError:
                raise UserValueError(
                    f"Failed to import user predict function "
                    f"'{self.user_predict_function_import_path}'. Please ensure that the path "
                    f"is available in the project directory '{self.project_directory}'."
                )
            except Exception:
                raise ArthurUserError(
                    f"Failed to import user predict function "
                    f"'{self.user_predict_function_import_path}'. Please check the predict "
                    f"function for errors."
                )
            load_image_func = (
                customer_model.load_image
                if self._model.input_type == InputType.Image
                else None
            )
            label_mapping = self.build_categorical_lookup_table()

            validate_predicted_attribute_order_matches_remote(self._model)

            self.explainer = ArthurExplainer(
                model_type=self._model.output_type,
                model_input_type=self._model.input_type,
                predicted_attributes=predicted_attrs,
                predict_func=customer_model.predict,
                data=explanation_training_data,
                categorical_features=categorical_attr_indices,
                label_mapping=label_mapping,
                shap_training_samples=self.shap_nrows,
                text_delimiter=self.text_delimiter,
                enable_shap=self.enable_shap,
                load_image_func=load_image_func,
                predictions_df=predictions_df,
                seed=self.seed,
            )
        finally:
            sys.path.remove(self.project_directory)
            os.chdir(current_dir)

    def get_request_files(self) -> List[Any]:
        """Returns files needed for request, for use with model.update_enrichments"""
        if self.zipfile is None:
            raise UnexpectedTypeError("zipfile is None")
        if self.project_directory is None:
            raise UnexpectedTypeError("project_directory is None")
        if self.requirements_file is None:
            raise UnexpectedTypeError("requirements_file is None")
        return [
            ("user_project.zip", open(self.zipfile, "rb")),
            (
                "user_requirements_file.txt",
                open(
                    os.path.join(self.project_directory, self.requirements_file), "rb"
                ),
                "text/plain",
            ),
            ("explainer.pkl", dill.dumps(self.explainer)),
        ]

    def get_request_config(self) -> Dict[str, str]:
        """Generates config dict for explainer. For use with model.update_enrichments"""
        config = {}

        # support shap not being enabled and not sending expected values
        if self.explainer is not None and self.explainer.enable_shap:
            expected_value = NumpyEncoder.convert_value(
                self.explainer.shap_expected_values()
            )
            if not isinstance(expected_value, list):
                expected_value = [expected_value]
            config["shap_expected_values"] = expected_value

        if self.python_version:
            config["python_version"] = self.python_version.split()[0]
        if self.sdk_version:
            config["sdk_version"] = self.sdk_version.split()[0]
        if self.user_predict_function_import_path:
            config["user_predict_function_import_path"] = (
                self.user_predict_function_import_path
            )
        if self.explanation_algo:
            config["explanation_algo"] = self.explanation_algo
        if self.streaming_explainability_enabled is not None:
            config["streaming_explainability_enabled"] = (
                self.streaming_explainability_enabled
            )
        if self.explanation_nsamples:
            config["explanation_nsamples"] = self.explanation_nsamples
        if self.model_server_num_cpu:
            config["model_server_cpu"] = self.model_server_num_cpu
        if self.model_server_memory:
            config["model_server_memory"] = self.model_server_memory
        if self.model_server_max_replicas:
            config["model_server_max_replicas"] = self.model_server_max_replicas
        if self.inference_consumer_num_cpu:
            config["inference_consumer_cpu"] = self.inference_consumer_num_cpu
        if self.inference_consumer_memory:
            config["inference_consumer_memory"] = self.inference_consumer_memory
        if self.inference_consumer_thread_pool_size:
            config["inference_consumer_thread_pool_size"] = (
                self.inference_consumer_thread_pool_size
            )
        if self.inference_consumer_score_percent:
            config["inference_consumer_score_percent"] = (
                self.inference_consumer_score_percent
            )
        if self.seed:
            config["seed"] = self.seed

        return config

    def make_zip(self) -> None:
        """Create the zip file for the project directory."""
        # ensure requirements file exists
        if self.project_directory is None:
            raise UnexpectedTypeError("project_directory is None")
        if self.requirements_file is None:
            raise UnexpectedTypeError("requirements_file is None")
        if not os.path.exists(
            os.path.join(self.project_directory, self.requirements_file)
        ):
            raise ArthurUserError(
                "File {0} does not exist".format(
                    os.path.join(self.project_directory, self.requirements_file)
                )
            )

        zipfile = PurePath(self.volume, "project").with_suffix(".zip")
        base_dir = PurePath(self.project_directory)

        if not os.path.exists(self.volume):
            os.makedirs(self.volume)

        # create a ZipFile object
        with ZipFile(zipfile, "w") as zip_file:
            # Iterate over all the files in directory
            for folder_name, sub_folders, filenames in os.walk(base_dir):
                print(f"Ignoring folder: {folder_name}")
                for filename in filenames:
                    file_path = PurePath(folder_name, filename)
                    relative_path = file_path.relative_to(base_dir)
                    parents_set = set(map(str, relative_path.parents))
                    if not parents_set.intersection(self.ignore_dirs):
                        zip_file.write(file_path, relative_path)

        self.zipfile = zipfile

    def test(self) -> None:
        """Download the docker container, install the files, and make sure the predict function can be run."""
        example = NumpyEncoder.convert_value(self.example)
        self.copy_files(example)

        volumes = {self.volume: {"bind": self.bind_path}}
        client = docker.from_env()
        print("Building docker image...")
        image = client.images.build(path=os.path.dirname(__file__), pull=True, rm=True)[
            0
        ]
        print("Creating container and testing predict function...")
        command = "{0}".format(os.path.join(self.bind_path, self.test_script))
        client.containers.run(image.id, volumes=volumes, command=command)
        print("Success!")

    def copy_files(self, example: Any) -> None:
        """Copy the files that are required for testing the docker image.

        :param: example: the example to use in the test script
        """

        explainer = os.path.join(self.volume, "explainer.pkl")
        dill.dump(self.explainer, open(explainer, "wb"))

        if self.project_directory is None:
            raise UnexpectedTypeError("project_directory is None")
        if self.requirements_file is None:
            raise UnexpectedTypeError("requirements_file is None")
        requirements = os.path.join(self.volume, self.requirements_file)
        shutil.copy2(
            os.path.join(self.project_directory, self.requirements_file), requirements
        )

        test = open(os.path.join(self.volume, self.test_script), "w")
        test.write(self.create_test_script(example))
        os.chmod(os.path.join(self.volume, self.test_script), 0o755)

    def create_test_script(self, example: Any) -> str:
        """Create a test script.

        :param: example:the example to use in the test script

        :return: the script contents
        """
        script = ";".join(
            [
                "from user_bundle.project import {0}".format(
                    self.user_predict_function_import_path
                ),
                "output = {0}.predict({1})".format(
                    self.user_predict_function_import_path, example
                ),
                "print(output)",
            ]
        )

        return "\n".join(
            [
                "#!/usr/bin/env bash",
                "set -e",
                'CODE_DIR="`pwd`/user_bundle"',
                'export USER_PROJECT_SOURCE_DIR="${CODE_DIR}/project"',
                'export USER_REQUIREMENTS_FILE="${CODE_DIR}/requirements.txt"',
                'export EXPLAINER_FILE="${CODE_DIR}/explainer.pkl"',
                'ENV_NAME="model-server"',
                "mkdir -p $CODE_DIR",
                "unzip -qq /mnt/{0}/project.zip -d $USER_PROJECT_SOURCE_DIR".format(
                    self.dirname
                ),
                "cp /mnt/{0}/{1} $USER_REQUIREMENTS_FILE".format(
                    self.dirname, self.requirements_file
                ),
                "cp /mnt/{0}/explainer.pkl $EXPLAINER_FILE".format(self.dirname),
                "conda init bash",
                "conda create --name $ENV_NAME python=$PYTHON_VERSION --quiet",
                "source ~/.bashrc",
                "source activate $ENV_NAME",
                "pip -qq install --no-cache-dir -r $USER_REQUIREMENTS_FILE",
                'python -c "{0}"'.format(script),
            ]
        )

    def validate_requirements_add_sklearn(self) -> List[str]:
        """Checks the user requirements file to make sure versions are correct.

        Also handles making sure scikit-learn version is pinned in user requirements
        file, if it is installed in local environment and not specified.

        scikit-learn is a dependency of shap, and arthurai doesn't require scikit-learn explicitly.
        This can lead to version A being installed and used locally when pickling explainer, then version B
        getting installed on model server as part of shap install. This can lead to pickle errors,
        since the pickled version of shap (client side) used a different version of sklearn than what is on model
        server.
        """
        if self.requirements_file is None:
            raise UnexpectedTypeError("requirements_file is None")
        with open(self.requirements_file, "r") as f:
            user_reqs = f.readlines()

        # build dict of installed packages and versions
        cur_env = {pkg.project_name: pkg.version for pkg in pkg_resources.working_set}

        errors = []

        # validate each line in user requirements file
        sklearn_found = False
        for req in user_reqs:
            req = req.replace("\n", "")
            # validate single pinned version: pandas==1.0.0
            if "==" not in req:
                errors.append(
                    f"Invalid line in requirements file: {req}. Each dependency must have a single specific "
                    f"version pinned using '=='"
                )
                continue
            splt = req.split("==")
            if len(splt) != 2:
                errors.append(
                    f"Invalid line in requirements file: {req}. Each dependency must have a single specific "
                    f"version pinned using '=='"
                )

            # flag if sklearn is already pinned in requirements file
            package, version = splt
            sklearn_found = package == "scikit-learn"

            # validate package version matches environment
            if package not in cur_env:
                logger.warn(
                    f"Explainability requirements file contains package not installed in current environment: "
                    f"{package}."
                )
            if package in cur_env and cur_env[package] != version:
                error_str = (
                    f"Explainability requirements file lists {package}=={version} but found version "
                    f"{cur_env[package]} currently installed."
                )
                error_str += " Packages listed in the requirements file must match the currently installed version."
                error_str += f" To resolve, update the requirements file to version {package}=={cur_env[package]}"
                errors.append(error_str)

        # insert sklearn into requirements file if necessary
        if "scikit-learn" in cur_env and not sklearn_found:
            cur_version = cur_env["scikit-learn"]
            if self.requirements_file is None:
                raise UnexpectedTypeError("requirements_file is None")
            with open(self.requirements_file, "a") as f:
                # ensure we add on new line
                leading_newline = "" if user_reqs[-1][-1] == "\n" else "\n"
                f.write(f"{leading_newline}scikit-learn=={cur_version}\n")
            logger.info(
                "scikit-learn was not included in explainer requirements file, but is installed in current "
                "environment. Automatically adding current version to requirements file. This helps ensure "
                "correct environment on model server."
            )

        return errors



import pytest
import responses

from arthurai import ArthurModel
from arthurai.common.constants import ModelStatus, ONBOARDING_SPINNER_MESSAGE, ONBOARDING_UPDATE_MESSAGE, InputType, \
    OutputType
from arthurai.common.exceptions import ArthurUnexpectedError
from arthurai.core.model_status_waiter import ModelStatusWaiter
from tests.helpers import mock_get
from tests.fixtures.models import regression as reg
from tests.fixtures.mocks import client

CREATING_STATUS_LIST = [ModelStatus.Ready, ModelStatus.CreationFailed]


def get_test_model(client):
    return ArthurModel(
        id=reg.TABULAR_MODEL_ID,
        partner_model_id="test",
        client=client.client,
        input_type=InputType.Tabular,
        output_type=OutputType.Multiclass,
    )


@responses.activate
def test_model_status_wait_creating_ready(client):
    model = get_test_model(client)
    mock_get(f"/api/v3/models/{reg.TABULAR_MODEL_ID}", response_body={
        'id': reg.TABULAR_MODEL_ID,
        'model_group_id': reg.TABULAR_MODEL_GROUP_ID,
        'version_sequence_num': 1,
        'attributes': [],
        'status': ModelStatus.Creating
    }, status=200)
    mock_get(f"/api/v3/models/{reg.TABULAR_MODEL_ID}", response_body={
        'id': reg.TABULAR_MODEL_ID,
        'model_group_id': reg.TABULAR_MODEL_GROUP_ID,
        'version_sequence_num': 1,
        'attributes': [],
        'status': ModelStatus.Ready
    }, status=200)

    model_status_waiter = ModelStatusWaiter(model, client=client.client)

    status = model_status_waiter.wait_for_valid_status(CREATING_STATUS_LIST, ONBOARDING_SPINNER_MESSAGE,
                                                       ONBOARDING_UPDATE_MESSAGE)
    assert len(responses.calls) == 2
    assert status == ModelStatus.Ready


@responses.activate
def test_model_status_wait_creating_failed(client):
    model = get_test_model(client)
    mock_get(f"/api/v3/models/{reg.TABULAR_MODEL_ID}", response_body={
        'id': reg.TABULAR_MODEL_ID,
        'model_group_id': reg.TABULAR_MODEL_GROUP_ID,
        'version_sequence_num': 1,
        'attributes': [],
        'status': ModelStatus.Creating
    }, status=200)
    mock_get(f"/api/v3/models/{reg.TABULAR_MODEL_ID}", response_body={
        'id': reg.TABULAR_MODEL_ID,
        'model_group_id': reg.TABULAR_MODEL_GROUP_ID,
        'version_sequence_num': 1,
        'attributes': [],
        'status': ModelStatus.CreationFailed
    }, status=200)

    model_status_waiter = ModelStatusWaiter(model, client=client.client)

    status = model_status_waiter.wait_for_valid_status(CREATING_STATUS_LIST, ONBOARDING_SPINNER_MESSAGE,
                                                       ONBOARDING_UPDATE_MESSAGE)
    assert len(responses.calls) == 2
    assert status == ModelStatus.CreationFailed


@responses.activate
def test_model_status_get_model_exception(client):
    model = get_test_model(client)
    mock_get(f"/api/v3/models/{reg.TABULAR_MODEL_ID}", response_body={}, status=500)

    model_status_waiter = ModelStatusWaiter(model, client=client.client)

    with pytest.raises(ArthurUnexpectedError):
        model_status_waiter.wait_for_valid_status(CREATING_STATUS_LIST, ONBOARDING_SPINNER_MESSAGE,
                                                  ONBOARDING_UPDATE_MESSAGE)
    assert len(responses.calls) == 1



from typing import NamedTuple, Dict, List, Optional, Union, Type

import logging
import pytest
import responses

from arthurai import ArthurModel
from arthurai.common.constants import ENRICHMENTS_SPINNER_MESSAGE, ENRICHMENTS_UPDATE_MESSAGE, \
    EnrichmentStatus, InputType, OutputType
from arthurai.common.exceptions import ArthurUnexpectedError, UserValueError
from arthurai.core.enrichment_status_waiter import EnrichmentStatusWaiter, await_enrichments_ready
from tests.helpers import mock_get
from tests.fixtures.mocks import client

READY_FOR_INFERENCES_STATUS_LIST = [EnrichmentStatus.EnrichmentStatusDisabled, EnrichmentStatus.EnrichmentStatusFailed,
                                    EnrichmentStatus.EnrichmentStatusTraining, EnrichmentStatus.EnrichmentStatusReady]

RESPONSE_NOT_READY = {
    "anomaly_detection": {
        "status": EnrichmentStatus.EnrichmentStatusDisabled
    },
    "explainability": {
        "status": EnrichmentStatus.EnrichmentStatusReady,
        "enabled": True
    },
    "hotspots": {
        "status": EnrichmentStatus.EnrichmentStatusPending,
    },
    "bias_mitigation": {
        "status": EnrichmentStatus.EnrichmentStatusTraining,
    }
}
RESPONSE_READY = {
    "anomaly_detection": {
        "status": EnrichmentStatus.EnrichmentStatusDisabled
    },
    "explainability": {
        "status": EnrichmentStatus.EnrichmentStatusReady,
        "enabled": True
    },
    "hotspots": {
        "status": EnrichmentStatus.EnrichmentStatusTraining,
    },
    "bias_mitigation": {
        "status": EnrichmentStatus.EnrichmentStatusFailed,
    }
}
RESPONSE_READY_WITHOUT_FAILED = {
    "anomaly_detection": {
        "status": EnrichmentStatus.EnrichmentStatusDisabled
    },
    "explainability": {
        "status": EnrichmentStatus.EnrichmentStatusReady,
        "enabled": True
    },
    "hotspots": {
        "status": EnrichmentStatus.EnrichmentStatusTraining,
    },
    "bias_mitigation": {
        "status": EnrichmentStatus.EnrichmentStatusReady,
    }
}

NO_STATUS_RESPONSE = {
    "explainability": {
        "enabled": True
    },
}


class ConstructEnrichmentStatusWaiterTestCase(NamedTuple):
    response_body: List[Dict[str, Dict]]
    status: List[int]
    expect_err: Optional[Union[str, Type[Exception]]]


WAIT_FOR_VALID_STATUS_CASES = [
    ConstructEnrichmentStatusWaiterTestCase(status=[200, 200], response_body=[RESPONSE_NOT_READY, RESPONSE_READY],
                                            expect_err=None),
    ConstructEnrichmentStatusWaiterTestCase(status=[500], response_body=[{}], expect_err="500 Internal Server Error"),
    ConstructEnrichmentStatusWaiterTestCase(status=[200], response_body=[NO_STATUS_RESPONSE],
                                            expect_err="ExpectedParameterNotFoundError")
]


def get_test_model(client) -> ArthurModel:
    return ArthurModel(
        id="test_model_id",
        partner_model_id="test",
        client=client.client,
        input_type=InputType.Tabular,
        output_type=OutputType.Multiclass,
    )


@pytest.mark.parametrize("case", WAIT_FOR_VALID_STATUS_CASES)
@responses.activate
def test_enrichment_status_wait_for_valid_status(client, case: ConstructEnrichmentStatusWaiterTestCase, caplog):
    model = get_test_model(client)
    for i in range(len(case.status)):
        mock_get(f"/api/v3/models/{model.id}/enrichments", response_body=case.response_body[i], status=case.status[i])

    enrichment_status_waiter = EnrichmentStatusWaiter(model)
    expected_statuses, statuses = [], []

    if case.expect_err is not None:
        with pytest.raises(ArthurUnexpectedError):
            with caplog.at_level(logging.ERROR):
                enrichments = enrichment_status_waiter.wait_for_valid_status(READY_FOR_INFERENCES_STATUS_LIST,
                                                                             ENRICHMENTS_SPINNER_MESSAGE,
                                                                             ENRICHMENTS_UPDATE_MESSAGE)
                statuses = [enrichment.status for enrichment in enrichments]
                assert len(statuses) == 0
        assert case.expect_err in caplog.text
    else:
        enrichments = enrichment_status_waiter.wait_for_valid_status(READY_FOR_INFERENCES_STATUS_LIST,
                                                                     ENRICHMENTS_SPINNER_MESSAGE,
                                                                     ENRICHMENTS_UPDATE_MESSAGE)
        statuses = [enrichment.status for enrichment in enrichments]
        for key, value in case.response_body[len(case.response_body) - 1].items():
            expected_statuses.append(value['status'])
        assert expected_statuses == statuses
    assert len(responses.calls) == len(case.status)


AWAIT_ENRICHMENTS_READY_CASES = [
    ConstructEnrichmentStatusWaiterTestCase(status=[200, 200],
                                            response_body=[RESPONSE_NOT_READY, RESPONSE_READY_WITHOUT_FAILED],
                                            expect_err=None),
    ConstructEnrichmentStatusWaiterTestCase(status=[200, 200], response_body=[RESPONSE_NOT_READY, RESPONSE_READY],
                                            expect_err=UserValueError),
    ConstructEnrichmentStatusWaiterTestCase(status=[500], response_body=[{}], expect_err=ArthurUnexpectedError),
    ConstructEnrichmentStatusWaiterTestCase(status=[200], response_body=[NO_STATUS_RESPONSE],
                                            expect_err=ArthurUnexpectedError)
]


@pytest.mark.parametrize("case", AWAIT_ENRICHMENTS_READY_CASES)
@responses.activate
def test_await_enrichments_ready(client, case: ConstructEnrichmentStatusWaiterTestCase):
    model = get_test_model(client)
    for i in range(len(case.status)):
        mock_get(f"/api/v3/models/{model.id}/enrichments", response_body=case.response_body[i], status=case.status[i])

    if case.expect_err is not None:
        with pytest.raises(case.expect_err):
            await_enrichments_ready(model)
    else:
        await_enrichments_ready(model)
    assert len(responses.calls) == len(case.status)



from typing import Optional
from datetime import datetime

from arthurai.common.exceptions import (
    arthur_excepted,
    UserValueError,
    MethodNotApplicableError,
)


class BiasMetrics(object):
    def __init__(self, arthur_model):
        self.model = arthur_model

    @arthur_excepted("Failed to get group-conditional confusion matrices")
    def group_confusion_matrices(
        self,
        attr_name: str,
        pred_property: Optional[str] = None,
        batch_id: Optional[str] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        return_by_metric=True,
    ):
        """
        Get group-conditional confusion matrices for all inferences, with the option to filter for a `batch_id` or
        a particular chunk of time. Currently only supports calculating bias with respect to a single sensitive
        attribute at a time. This method handles both binary and multiclass models.

        :param attr_name: The name of the sensitive attribute to get metrics for.
        :param pred_property: For multiclass models, the predicted label to get the confusion matrix for.
        :param batch_id: Optional filter to limit calculations to a specific batch ID.
        :param start_time: Optional filter to limit calculations to inferences after a specific timestamp.
        :param end_time: Optional filter to limit calculations to inferences before a specific timestamp.
        :param return_by_metric: Whether the returned dictionary of results should be keyed by metric or by sensitive attribute value.

        :return: Either a dict of the form `{ metric: { sens_1: val1, sens_2: val2 }, ...}` (default) or its reverse.
        """

        # check whether model has any attributes marked as `monitor_for_bias`
        if not self.model.check_has_bias_attrs():
            raise MethodNotApplicableError(
                "This model does not have any attributes set to monitor for bias."
            )

        if not self.model.check_attr_is_bias(attr_name):
            raise MethodNotApplicableError(
                "This attribute is not set to monitor for bias."
            )

        if batch_id and not self.model.is_batch:
            raise UserValueError(
                "This is not a batch model; remove the `batch_id` argument."
            )

        pos_attr = self.model.get_positive_predicted_class()

        if pos_attr:
            q = _bias_cm_query(
                attr_name,
                batch_id=batch_id,
                classifier_threshold=self.model.classifier_threshold,
                start_time=start_time,
                end_time=end_time,
            )
        elif not pred_property:
            raise UserValueError(
                "For a multiclass model, the `pred_property` parameter must be specified."
            )
        else:
            q = _bias_multiclass_cm_query(
                attr_name,
                pred_property=pred_property,
                batch_id=batch_id,
                start_time=start_time,
                end_time=end_time,
            )

        resp = self.model.query(q)

        sens_keyed = {x[attr_name]: x["confusion_matrix"] for x in resp}

        if return_by_metric:
            metr_keys = [
                "accuracy_rate",
                "balanced_accuracy_rate",
                "f1",
                "false_negative_rate",
                "false_positive_rate",
                "precision",
                "true_negative_rate",
                "true_positive_rate",
            ]
            return {
                x: {s: sens_keyed[s][x] for s in sens_keyed.keys()} for x in metr_keys
            }  # reverse the dictionary

        return sens_keyed

    @arthur_excepted("Failed to get group-conditional positivity rates")
    def group_positivity_rates(
        self,
        attr_name: str,
        batch_id: Optional[str] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
    ):
        """
        Get group-conditional positivity rates for all inferences, with the option to filter for a `batch_id` or
        a particular chunk of time. Currently only supports calculating bias with respect to a single sensitive
        attribute at a time. This method currently only handles binary models.

        :param attr_name: The name of the sensitive attribute to get metrics for.
        :param batch_id: Optional filter to limit calculations to a specific batch ID.
        :param start_time: Optional filter to limit calculations to inferences after a specific timestamp.
        :param end_time: Optional filter to limit calculations to inferences before a specific timestamp.

        :return: A dict of the form {sens_1: pos_1, sens_2: pos_2, ...}
        """

        # check whether model has any attributes marked as `monitor_for_bias`
        if not self.model.check_has_bias_attrs():
            raise MethodNotApplicableError(
                "This model does not have any attributes set to monitor for bias."
            )

        if not self.model.check_attr_is_bias(attr_name):
            raise MethodNotApplicableError(
                "This attribute is not set to monitor for bias."
            )

        if batch_id and not self.model.is_batch:
            raise UserValueError(
                "This is not a batch model; remove the `batch_id` argument."
            )

        pos_attr = self.model.get_positive_predicted_class()

        if pos_attr:
            q = _bias_pr_query(
                attr_name,
                pred_property=pos_attr,
                batch_id=batch_id,
                classifier_threshold=self.model.classifier_threshold,
                start_time=start_time,
                end_time=end_time,
            )
        else:
            raise UserValueError(
                "Positivity rates are currently only supported for binary classifiers."
            )

        resp = self.model.query(q)
        return {x[attr_name]: x["positive_rate"] for x in resp}

    # ALIAS METHOD NAMES
    demographic_parity = group_positivity_rates


def _bias_cm_query(
    sens_attr, classifier_threshold=0.5, batch_id=None, start_time=None, end_time=None
):
    """
    Get group-conditional `metric` (either `confusionMatrixRate` or `rate`) for BINARY classifier
    If batch_id or start/end times are None, then do this over all batches / all time
    Query is identical for streaming
    """

    query = {
        "select": [
            {"property": sens_attr},
            {
                "function": "confusionMatrixRate",
                "alias": "confusion_matrix",
                "parameters": {"threshold": classifier_threshold},
            },
        ],
        "group_by": [{"property": sens_attr}],
    }

    query = _add_filters(query, batch_id, start_time, end_time)

    return query


def _bias_multiclass_cm_query(
    sens_attr, pred_property, batch_id=None, start_time=None, end_time=None
):
    """
    Get group-conditional confusion matrix for MULTICLASS classifier (`rate` doesn't apply here, since there is no single threshold)
    If batch_id or start/end times are None, then do this over all batches / all time
    Query is identical for streaming
    """

    query = {
        "select": [
            {"property": sens_attr},
            {
                "function": "confusionMatrixRateMulticlass",
                "alias": "confusion_matrix",
                "parameters": {"predicted_property": pred_property},
            },
        ],
        "group_by": [{"property": sens_attr}],
    }

    query = _add_filters(query, batch_id, start_time, end_time)

    return query


def _bias_pr_query(
    sens_attr,
    pred_property,
    classifier_threshold=0.5,
    batch_id=None,
    start_time=None,
    end_time=None,
):
    """
    Get group-conditional positivity rates for BINARY classifier
    If batch_id or start/end times are None, then do this over all batches/all time
    Query is identical for streaming
    """

    query = {
        "select": [
            {"property": sens_attr},
            {
                "function": "rate",
                "alias": "positive_rate",
                "parameters": {
                    "property": pred_property,
                    "comparator": "gte",
                    "value": classifier_threshold,
                },
            },
        ],
        "group_by": [{"property": sens_attr}],
    }

    query = _add_filters(query, batch_id, start_time, end_time)

    return query


def _add_filters(query, batch_id=None, start_time=None, end_time=None):
    if batch_id:
        query["filter"] = [
            {"property": "batch_id", "comparator": "eq", "value": batch_id}
        ]

    if start_time:
        query["filter"] = [
            {
                "property": "inference_timestamp",
                "comparator": "gte",
                "value": start_time.strftime("%Y-%m-%dT%H:%M:%SZ"),
            }
        ]

    if end_time:
        query["filter"] = [
            {
                "property": "inference_timestamp",
                "comparator": "lte",
                "value": end_time.strftime("%Y-%m-%dT%H:%M:%SZ"),
            }
        ]

    return query



import datetime

import numpy as np
import pandas as pd
import pytest
import pytz

from arthurai import ArthurAttribute, ArthurModel
from arthurai.common.constants import ValueType, Stage, OutputType, InputType, InferenceType

from tests.fixtures.mocks import client

GROUND_TRUTH = "gt"
TEMPO = "tempo"
PRED = "recommendations"
TIME_SERIES = "bill_pay_history"
TIME = "time"

RANKED_LIST_DATAFRAME = pd.DataFrame({
    TIME_SERIES: [[{"timestamp": datetime.datetime(2011, 8, 15, 0, 0, 0, 0, pytz.UTC), "value": 1},
                   {"timestamp": datetime.datetime(2011, 8, 16, 0, 0, 0, 0, pytz.UTC), "value": 2},
                   {"timestamp": datetime.datetime(2011, 8, 17, 0, 0, 0, 0, pytz.UTC), "value": 3}],
                  [{"timestamp": datetime.datetime(2011, 8, 25, 0, 0, 0, 0, pytz.UTC), "value": 3},
                   {"timestamp": datetime.datetime(2011, 8, 26, 0, 0, 0, 0, pytz.UTC), "value": 2},
                   {"timestamp": datetime.datetime(2011, 8, 27, 0, 0, 0, 0, pytz.UTC), "value": 1}]],
    TIME: ["2011-08-25T00:00:00Z", "2011-08-28T00:00:00Z",],
    TEMPO: [88, 66],
    PRED: [
        [{"item_id": "1", "label": "test", "score": 0.88},
         {"item_id": "2", "label": "test", "score": 0.88}],
        [{"item_id": "1", "label": "test", "score": 0.66},
         {"item_id": "2", "label": "test", "score": 0.66}],
    ],
    GROUND_TRUTH: [["test1", "test2"], ["test3", "test4"]]
})

RANKED_LIST_JSON = [
    {
        InferenceType.REFERENCE_DATA: {
            TIME_SERIES: [
                {"timestamp": "2011-08-15T00:00:00+00:00", "value": 1},
                {"timestamp": "2011-08-16T00:00:00+00:00", "value": 2},
                {"timestamp": "2011-08-17T00:00:00+00:00", "value": 3}
            ],
            TIME: "2011-08-25T00:00:00Z",
            TEMPO: 88,
            PRED: [
                {"item_id": "1", "label": "test", "score": 0.88},
                {"item_id": "2", "label": "test", "score": 0.88}
            ],
            GROUND_TRUTH: ["test1", "test2"],
        }
    },
    {
        InferenceType.REFERENCE_DATA: {
            TIME_SERIES: [
                {"timestamp": "2011-08-25T00:00:00+00:00", "value": 3},
                {"timestamp": "2011-08-26T00:00:00+00:00", "value": 2},
                {"timestamp": "2011-08-27T00:00:00+00:00", "value": 1}
            ],
            TIME: "2011-08-28T00:00:00Z",
            TEMPO: 66,
            PRED: [
                {"item_id": "1", "label": "test", "score": 0.66},
                {"item_id": "2", "label": "test", "score": 0.66}
            ],
            GROUND_TRUTH: ["test3", "test4"]
        }
    }
]

TEMPO_ATTR = ArthurAttribute(name=TEMPO,
                             value_type=ValueType.Integer,
                             stage=Stage.NonInputData,
                             position=0)

TIME_ATTR = ArthurAttribute(name=TIME,
                            value_type=ValueType.Timestamp,
                            stage=Stage.NonInputData,
                            position=1)

TIME_SERIES_ATTR = ArthurAttribute(name=TIME_SERIES,
                                   value_type=ValueType.TimeSeries,
                                   stage=Stage.ModelPipelineInput,
                                   position=0)

GT_ATTR = ArthurAttribute(
            name=GROUND_TRUTH,
            stage=Stage.GroundTruth,
            value_type=ValueType.StringArray,
            attribute_link=PRED,
            position=0)

PRED_ATTR = ArthurAttribute(
            name=PRED,
            stage=Stage.PredictedValue,
            value_type=ValueType.RankedList,
            attribute_link=GROUND_TRUTH,
            position=0)

RANKED_LIST_MODEL_ATTRIBUTES = [TEMPO_ATTR, TIME_ATTR, GT_ATTR, PRED_ATTR, TIME_SERIES_ATTR]

MODEL_DATA = {"partner_model_id": "test",
              "input_type": InputType.TimeSeries,
              "output_type": OutputType.RankedList}

MODEL_ID = "c4ea58b6-4ec7-43b2-94d3-786eccb2a492"


@pytest.fixture
def ranked_list_model(client):
    model = ArthurModel(
        **MODEL_DATA,
        client=client.client,
        attributes=RANKED_LIST_MODEL_ATTRIBUTES
    )
    model.id = MODEL_ID
    return model



import pandas as pd

from arthurai.common.constants import ValueType
from arthurai.core.attributes import AttributeCategory
from arthurai.core.models import ArthurModel


class TestModelsServiceObject:

    def test_get_attribute_data(self):
        """
        Currently, tests that float attributes that only take on values of
        0.0 or 1.0 are converted to categorical.
        """

        value_type = ValueType.Float

        df = pd.DataFrame({
            'float_binary': [0.0, 1.0, 1.0, 1.0, 0.0],
            'true_float': [0.0, 0.2, 0.4, 0.6, 0.8]
        })

        binary_attr = ArthurModel._get_attribute_data(value_type, df.float_binary)
        float_attr = ArthurModel._get_attribute_data(value_type, df.true_float)

        assert binary_attr["categorical"] is True
        assert float_attr["categorical"] is False

    def test_unstructured_text_get_attribute_data(self):
        """
        Ensure non-unique unstructured text attributes don't get
        categories
        """
        value_type = ValueType.Unstructured_Text
        df = pd.DataFrame({
            "text": ["test", "test", "more", "things"],
            "target": [0,1,2,3]
        })

        text_attr = ArthurModel._get_attribute_data(value_type, df.text)

        assert text_attr['is_unique'] == False
        assert text_attr['categorical'] == True
        assert 'categories' not in text_attr

    def test_categories_ordering_get_attribute_data(self):
        df_float = pd.DataFrame({
            'float_binary': [0.0, 1.0, 1.0, 1.0, 0.0],
        })

        df_string = pd.DataFrame({
            'strings': ["five", "two", "six", "three", "three", "four", "four", "one", "four", "four", "five",
                        "five", "six", "five", "five", "six", "three", "six", "six", "two", "six"]
        })

        float_ordering = [AttributeCategory(value=1.0), AttributeCategory(value=0.0)]
        strings_ordering = [AttributeCategory(value="six"), AttributeCategory(value="five"),
                            AttributeCategory(value="four"), AttributeCategory(value="three"),
                            AttributeCategory(value="two"), AttributeCategory(value="one")]

        binary_attr = ArthurModel._get_attribute_data(ValueType.Float, df_float.float_binary)
        string_attr = ArthurModel._get_attribute_data(ValueType.String, df_string.strings)

        assert binary_attr["categories"] == float_ordering
        assert string_attr["categories"] == strings_ordering



import logging
from typing import List

from arthurai.common.constants import (
    EnrichmentStatus,
    ENRICHMENT_READY_OR_FAILED_STATES,
    ENRICHMENTS_SPINNER_MESSAGE,
    ENRICHMENTS_UPDATE_MESSAGE,
)
from arthurai.common.exceptions import ExpectedParameterNotFoundError, UserValueError
from arthurai.core.status_waiter import StatusWaiter

logger = logging.getLogger(__name__)


class StatusForEnrichment:
    def __init__(self, enrichment_name: str, status: EnrichmentStatus):
        self.enrichment_name = enrichment_name
        self.status = status


class EnrichmentStatusWaiter(StatusWaiter):
    def __init__(self, model):
        self.err_mess = (
            "Failed when trying to fetch the statuses of the enrichments. "
            "Please retry to wait for the statuses of the enrichments."
        )
        super().__init__(model, self.err_mess)

    def _check_valid_status(self, status_list: List) -> bool:
        """Returns True if all enrichment statuses have reached valid final states in `status_list`, False otherwise"""
        for enrichment in self._status:
            if enrichment.status not in status_list:
                return False
        return True

    def _update_status(self):
        """Updates `_status` with the latest enrichment statuses"""
        self._status = []
        enrichments = self.model.get_enrichments()
        for key, val in enrichments.items():
            if "status" not in val:
                raise ExpectedParameterNotFoundError(
                    f"An error occurred: 'status' was not found in enrichment "
                    f"{key}—got {val} instead"
                )
            else:
                self._status.append(StatusForEnrichment(key, val["status"]))


def await_enrichments_ready(model):
    """wait for enrichments to be Ready, Training, Disabled, or Failed

    Wait until all enrichments for the model reach
    :py:attr:`~arthurai.common.constants.EnrichmentStatus.EnrichmentStatusReady`,
    :py:attr:`~arthurai.common.constants.EnrichmentStatus.EnrichmentStatusTraining`,
    :py:attr:`~arthurai.common.constants.EnrichmentStatus.EnrichmentStatusFailed`, or
    :py:attr:`~arthurai.common.constants.EnrichmentStatus.EnrichmentStatusDisabled` states. Prints to the log that it is
    waiting for the enrichments to be ready to receive inferences until all enrichments reach valid enrichment statuses.

    :raises UserValueError: If an exception occurs when checking the status of the model's enrichments or all
        enrichments have reached valid statuses and at least one enrichment is in the
        :py:attr:`~arthurai.common.constants.EnrichmentStatus.EnrichmentStatusFailed` state.
    """
    enrichment_statuses = EnrichmentStatusWaiter(model).wait_for_valid_status(
        ENRICHMENT_READY_OR_FAILED_STATES,
        ENRICHMENTS_SPINNER_MESSAGE,
        ENRICHMENTS_UPDATE_MESSAGE,
    )
    failed_workflows = []
    for enrichment in enrichment_statuses:
        if enrichment.status == EnrichmentStatus.EnrichmentStatusFailed:
            failed_workflows.append(enrichment.enrichment_name)
    if len(failed_workflows) != 0:
        raise UserValueError(
            f"The following enrichments failed to provision: {', '.join(failed_workflows)}. Please "
            f"fix the failed enrichments before sending inferences, or the enrichments will not "
            f"process correctly. If you wish to proceed without the enrichments, set the "
            f"wait_for_enrichments parameter to False when calling the 'send_bulk_inferences' or "
            f"'send_inferences' functions."
        )
    else:
        logger.info(
            "All enrichments are ready to accept inferences. Continuing with inference upload."
        )



import logging
import threading
import time
from abc import ABC, abstractmethod
from typing import List, Any

from yaspin import yaspin

from arthurai.client.http.requests import HTTPClient
from arthurai.common.exceptions import ArthurUnexpectedError

logger = logging.getLogger(__name__)


class StatusWaiter(ABC):
    def __init__(self, model, err_mess: str, client: HTTPClient = None):
        """Superclass for waiting on some attribute of a model to reach a final state

        See :class:`arthurai.core.ModelStatusWaiter` for an example subclass implementation.

        :param model: ArthurModel for which the status of some related attribute is being waited on
        :param err_mess: Error message to log when the `update_status_func` function raises an exception.
        :param client: Optional HTTPClient.
        """
        self.model = model
        self._client = client
        self._status = None
        self._status_exception: bool = False
        self._status_available = threading.Event()
        self.err_mess = err_mess

    def wait_for_valid_status(
        self,
        status_list: List,
        spinner_message: str,
        update_message: str,
        update_interval_seconds=60,
    ) -> Any:
        """Wait to reach a status on the status_list

        Starts the thread that is responsible for checking the status until the attribute being checked reaches a status
        passed in the list as an argument or an exception occurs, and runs a spinner providing updates while waiting for
        the valid final status state.

        :param status_list: List of final valid statuses.
        :param spinner_message: The message you want to display when spinner is running and waiting for results
        :param update_message: The message you want to display when update_interval_seconds time is exceeded
        :param update_interval_seconds: Frequency at which an update is provided on the console
        """

        self._status_available.clear()
        self._status_exception = False
        thread = threading.Thread(
            target=self._await_final_status_thread, args=[status_list]
        )
        thread.start()

        with yaspin(text=spinner_message) as sp:
            while not self._status_available.wait(timeout=update_interval_seconds):
                sp.write(update_message)
            sp.stop()

        if self._status_exception is True:
            raise ArthurUnexpectedError(self.err_mess)

        return self._status

    def _await_final_status_thread(self, status_list: List, poll_interval_seconds=5):
        """Thread that polls the current status

        Polls the current status until it reaches a status in the status_list or an exception occurs.
        This method is used as a thread and sets the flag _status_available at the end of execution.
        Do not use as a regular function.
        :param status_list: List of valid statuses
        :param poll_interval_seconds: polling interval to check status on the model
        """
        while True:
            try:
                self._update_status()
            except BaseException as ex:
                logger.exception(ex)
                self._status_exception = True
                self._status_available.set()  # setting the flag on the threading event when result is available
                break
            if self._check_valid_status(
                status_list
            ):  # if the final state is reached, set the flag and break
                self._status_available.set()
                break
            time.sleep(poll_interval_seconds)

    @abstractmethod
    def _check_valid_status(self, status_list: List) -> bool:
        """Returns a boolean indicating if a final valid status has been reached.

        Must be overridden by subclass.

        :param status_list: List of final valid statuses.
        """
        raise NotImplementedError

    @abstractmethod
    def _update_status(self):
        """Updates the `StatusWaiter._status` field with the current status of the attribute being waited on.

        Must be overridden by subclass.
        """
        raise NotImplementedError



import logging
from http import HTTPStatus
from typing import List

from arthurai.client.http.requests import HTTPClient
from arthurai.common.exceptions import ExpectedParameterNotFoundError
from arthurai.core.status_waiter import StatusWaiter

logger = logging.getLogger(__name__)


class ModelStatusWaiter(StatusWaiter):
    def __init__(self, model, client: HTTPClient):
        self.err_mess = (
            "Failed when trying to fetch the status of the model. "
            "Please retry to wait for the status of the model."
        )
        super().__init__(model, self.err_mess, client)

    def _update_status(self):
        """Updates `_status` with the latest status of the current model"""
        model_resp = self._client.get(
            f"/models/{self.model.id}",
            return_raw_response=True,
            validation_response_code=HTTPStatus.OK,
        )
        model_resp_json = model_resp.json()
        if "id" not in model_resp_json:
            raise ExpectedParameterNotFoundError(
                f"An error occurred: {model_resp}, {model_resp.status_code}, {model_resp.content}"
            )
        self._status = model_resp_json["status"]

    def _check_valid_status(self, status_list: List) -> bool:
        """Returns True if model status has reached valid final state in `status_list`, False otherwise"""
        return self._status in status_list



# arthur_example.py
from dataclasses import dataclass
import pandas as pd
from pathlib import Path
from sklearn.model_selection import train_test_split
from typing import List, Union, Optional, Tuple
from arthurai.core.base import ArthurBaseJsonDataclass
from arthurai.core.models import ArthurModel
from arthurai.common.constants import Stage
from arthurai.common.exceptions import ArthurDownloadError
from arthurai.datasets.download import (
    ArthurExampleDownloader,
    default_download_directory,
    load_downloaded_file,
)


@dataclass
class ArthurExampleSchema(ArthurBaseJsonDataclass):
    """
    Class for the schema of all the data used in an Arthur Example
    """

    arthur_model: ArthurModel
    default_data_split_test_size: Optional[float] = 0.3
    default_random_state: int = 278487


class ArthurExample:
    """
    Class for a user to interface with example data for model analysis and using Arthur
    """

    def __init__(self, name: str, download_destination_folder: Optional[Path] = None):
        """
        Constructs an ArthurExample

        :param name: str, the name of the example
        :param download_destination_folder: Path, the destination to save files to. Defaults to current directory if
                none is specified.
        """
        if download_destination_folder is None:
            download_destination_folder = Path.cwd() / default_download_directory
        try:
            downloader = ArthurExampleDownloader(name, download_destination_folder)
        except ArthurDownloadError:
            raise ValueError(
                f"Cannot download data: {name} is not the name of a valid ArthurExample."
            ) from None

        self.example_schema = ArthurExampleSchema.from_dict(
            downloader.download_metadata()
        )
        self.dataset = load_downloaded_file(downloader.download_dataset())
        self.predictions = load_downloaded_file(
            downloader.download_pretrained_model_predictions()
        )

    def get_dataset(
        self, split: bool = False, test_split_size: Optional[float] = None
    ) -> Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame]]:
        """
        Returns a dataframe of the model's inputs, predictions, ground truth labels, and non input data

        :param split: bool, whether to return the data split into train/test.
        :param test_split_size: the percentage of data to be split into the test set if splitting the data. If None,
                uses the default test split defined in the example schema.
        :return: dataframe of inputs, predictions, ground truth labels, and non input data
        """
        if split:
            train_inputs, test_inputs = self.get_inputs(
                split=True, test_split_size=test_split_size
            )
            train_pred, test_pred = self.get_predictions(
                split=True, test_split_size=test_split_size
            )
            train_gt, test_gt = self.get_ground_truth_data(
                split=True, test_split_size=test_split_size
            )
            train_dataset = pd.concat([train_inputs, train_pred, train_gt], axis=1)
            test_dataset = pd.concat([test_inputs, test_pred, test_gt], axis=1)
            if len(self.non_input_attribute_names) > 0:
                train_non_input, test_non_input = self.get_non_input_data(
                    split=True, test_split_size=test_split_size
                )
                train_dataset = pd.concat([train_dataset, train_non_input], axis=1)
                test_dataset = pd.concat([test_dataset, test_non_input], axis=1)
            return train_dataset, test_dataset
        else:
            dataset = pd.concat(
                [
                    self.get_inputs(),
                    self.get_predictions(),
                    self.get_ground_truth_data(),
                ],
                axis=1,
            )
            if len(self.non_input_attribute_names) > 0:
                dataset = pd.concat([dataset, self.get_non_input_data()], axis=1)
            return dataset

    def get_inputs(
        self,
        split: bool = False,
        test_split_size: Optional[float] = None,
        random_state: Optional[int] = None,
    ) -> Union[
        pd.DataFrame,
        pd.Series,
        Tuple[pd.DataFrame, pd.DataFrame],
        Tuple[pd.Series, pd.Series],
    ]:
        """
        Returns the model's input attribute feature values from self.dataset

        :param split: bool, whether to return the data split by reference/inference (AKA train/test),
                or whether to return a single dataframe
        :param test_split_size: the percentage of data to be split into the test set if splitting the data
        :param random_state: int, random state for optional split
        :return: dataframe of model input features
        """
        X = self.dataset[self.input_attribute_names]
        if split:
            if test_split_size is None:
                test_split_size = self.example_schema.default_data_split_test_size
            if random_state is None:
                random_state = self.example_schema.default_random_state
            return train_test_split(
                X, test_size=test_split_size, random_state=random_state
            )
        return X

    def get_ground_truth_data(
        self,
        split: bool = False,
        test_split_size: Optional[float] = None,
        random_state: Optional[int] = None,
    ) -> Union[
        pd.DataFrame,
        pd.Series,
        Tuple[pd.DataFrame, pd.DataFrame],
        Tuple[pd.Series, pd.Series],
    ]:
        """
        Returns the model's ground truth labels from self.dataset

        :param split: bool, whether to return the data split by reference/inference (AKA train/test),
                or whether to return a single dataframe
        :param test_split_size: the percentage of data to be split into the test set if splitting the data
        :param random_state: int, random state for optional split
        :return: dataframe of model ground truth values
        """
        y = self.dataset[self.gt_attribute_names]
        if split:
            if test_split_size is None:
                test_split_size = self.example_schema.default_data_split_test_size
            if random_state is None:
                random_state = self.example_schema.default_random_state
            return train_test_split(
                y, test_size=test_split_size, random_state=random_state
            )
        return y

    def get_predictions(
        self,
        split: bool = False,
        test_split_size: Optional[float] = None,
        random_state: Optional[int] = None,
    ) -> Union[
        pd.DataFrame,
        pd.Series,
        Tuple[pd.DataFrame, pd.DataFrame],
        Tuple[pd.Series, pd.Series],
    ]:
        """
        Returns the model's predicted values from self.predictions

        :param split: bool, whether to return the data split by reference/inference (AKA train/test),
                or whether to return a single dataframe
        :param test_split_size: the percentage of data to be split into the test set if splitting the data
        :param random_state: int, random state for optional split
        :return: dataframe of model prediction data
        """
        y_pred = self.predictions[self.pred_attribute_names]
        if split:
            if test_split_size is None:
                test_split_size = self.example_schema.default_data_split_test_size
            if random_state is None:
                random_state = self.example_schema.default_random_state
            return train_test_split(
                y_pred, test_size=test_split_size, random_state=random_state
            )
        return y_pred

    def get_non_input_data(
        self,
        split: bool = False,
        test_split_size: Optional[float] = None,
        random_state: Optional[int] = None,
    ) -> Union[
        pd.DataFrame,
        pd.Series,
        Tuple[pd.DataFrame, pd.DataFrame],
        Tuple[pd.Series, pd.Series],
    ]:
        """
        Returns the model's non input attributes from self.dataset

        :param split: bool, whether to return the data split by reference/inference (AKA train/test),
                or whether to return a single dataframe
        :param test_split_size: the percentage of data to be split into the test set if splitting the data
        :param random_state: int, random state for optional split
        :return: dataframe of model non-input attribute data
        """
        if len(self.non_input_attribute_names) > 0:
            non_input_data = self.dataset[self.non_input_attribute_names]
            if split:
                if test_split_size is None:
                    test_split_size = self.example_schema.default_data_split_test_size
                if random_state is None:
                    random_state = self.example_schema.default_random_state
                return train_test_split(
                    non_input_data, test_size=test_split_size, random_state=random_state
                )
            return non_input_data
        raise ValueError("This example has no non_input attributes.")

    @property
    def input_attribute_names(self) -> List[str]:
        """
        The list of feature names which make up the model's input attributes

        :return: list of input attribute names
        """
        return [
            attr.name
            for attr in self.example_schema.arthur_model.get_attributes(
                stage=Stage.ModelPipelineInput
            )
        ]

    @property
    def gt_attribute_names(self) -> List[str]:
        """
        The list of attribute names which make up the model's ground truth attributes
        Either GroundTruth or GroundTruthClass

        :return: list of ground truth attribute names
        """
        gt_attrs = [
            attr.name
            for attr in self.example_schema.arthur_model.get_attributes(
                stage=Stage.GroundTruth
            )
        ]
        if len(gt_attrs) > 0:
            return gt_attrs
        return [
            attr.name
            for attr in self.example_schema.arthur_model.get_attributes(
                stage=Stage.GroundTruthClass
            )
        ]

    @property
    def pred_attribute_names(self) -> List[str]:
        """
        The list of feature names which make up the model's predicted attributes

        :return: list of predicted attribute names
        """
        return [
            attr.name
            for attr in self.example_schema.arthur_model.get_attributes(
                stage=Stage.PredictedValue
            )
        ]

    @property
    def non_input_attribute_names(self) -> List[str]:
        """
        The list of feature names which make up the model's non-input attributes

        :return: list of non-input attribute names
        """
        return [
            attr.name
            for attr in self.example_schema.arthur_model.get_attributes(
                stage=Stage.NonInputData
            )
        ]



import pytest
from pytest import approx
import numpy as np
from arthurai.core.model_utils import tensors_to_arthur_inference

VOCAB = {
    0: "this",
    1: "is",
    2: "test"
}


@pytest.mark.parametrize('num_probs,expected_formatted', [
    (1, [approx({"this": 0.7}), approx({"test": 0.99}), approx({"test": 0.3333})]),
    (2, [approx({"this": 0.7, "is": 0.2}), approx({"test": 0.99, "is": 0.005}), approx({"test": 0.3333, "is": 0.3333})]),
    (3, [approx({"this": 0.7, "is": 0.2, "test": 0.1}), approx({"test": 0.99, "is": 0.005, "this": 0.005}),
         approx({"this": 0.3333, "is": 0.3333, "test": 0.3333})])
])
def test_tensors_to_arthur_inference(num_probs, expected_formatted):
    log_tensor = np.array([[-0.35667494, -1.60943791, -2.30258509],
                           [-5.29831737, -5.29831737, -0.01005034],
                           [-1.09871229, -1.09871229, -1.09871229]])
    arthur_likelihoods = tensors_to_arthur_inference(log_tensor, VOCAB, num_probs)
    assert arthur_likelihoods == expected_formatted


@pytest.mark.parametrize('log_tensor,num_probs', [
    (np.array([[-0.35667494, -1.2039728, -2.30258509]]), 2),
    (np.array([[-0.35667494, -1.60943791, -2.30258509]]), 6),
    (np.array([[[[1]], [[2]], [[3]]]]), 2)
], ids=['invalid probs', 'num probs > 5', 'invalid shape'])
def test_tensors_to_arthur_inference_invalid(log_tensor, num_probs):
    with pytest.raises(ValueError):
        tensors_to_arthur_inference(log_tensor, VOCAB, num_probs)


import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import Any, List, Union

from arthurai.common.exceptions import arthur_excepted
from arthurai.core.viz import style, utils
import warnings

warnings.filterwarnings("ignore")


class DataVisualizer(object):
    def __init__(self, arthur_models):
        self.models = arthur_models

    @arthur_excepted("failed to generate timeline")
    def timeline(self, attribute_name):
        """Generates a visualization of the distribution of an attribute over time.
        For categorical attributes, a stacked area chart over time.
        For continuous attributes, a joyplot showing probability densities over time.
        :raises ArthurUserError: failed due to user error
        :raises ArthurUnexpectedError: failed due to an unexpected error
        """
        for model in self.models:
            if model.get_attribute(attribute_name).categorical:
                self._timeline_categorical(model, attribute_name)
            else:
                self._timeline_continuous(model, attribute_name)

    @arthur_excepted("failed to generate metric series")
    def metric_series(self, metric_names, time_resolution="day"):
        """Generates a line series visualization for selected metrics.

        .. code-block:: python

            # plot both the model's AUC metric and the model's FPR metric by the hour
            arthur_model.viz.metric_series(["auc", "falsePositiveRate"], time_resolution="hour")

        :raises ArthurUserError: failed due to user error
        :raises ArthurUnexpectedError: failed due to an unexpected error
        """
        fig = plt.figure(figsize=(20, 4))
        ax = fig.add_subplot(1, 1, 1)
        predicted_property, ground_truth_property = utils.get_pred_and_gt_attrs(
            self.models[0]
        )
        pal = style.categorical_palette

        # Define list of palettes to be used
        num_metrics = len(metric_names)
        max_metrics = len(style.multi_categorical_palette_names)
        if num_metrics > max_metrics:
            raise UserValueError(
                f"Too many metrics to display. The maximum is {max_metrics}, you wanted to see {num_metrics}"
            )
        color_palette_names = style.multi_categorical_palette_names[0:num_metrics]

        for metric_index, metric_name in enumerate(metric_names):
            color_palette = sns.color_palette(
                palette=color_palette_names[metric_index], n_colors=len(self.models)
            )
            for model_index, model in enumerate(self.models):
                if model.is_batch:
                    time_resolution = "batch_id"
                    query = _metric_series_batch_query(
                        metric_name,
                        ground_truth_property,
                        predicted_property,
                        model.classifier_threshold,
                    )
                else:
                    query = _metric_series_streaming_query(
                        metric_name,
                        time_resolution,
                        ground_truth_property,
                        predicted_property,
                        model.classifier_threshold,
                    )

                response = model.query(query)
                df = pd.DataFrame(response)[::-1]

                label = _label(model, metric_name)
                color = _alpha_color(color_palette[model_index])
                plt.plot(
                    df[time_resolution],
                    df[metric_name],
                    lw=12,
                    color=color,
                    label=label,
                )
                plt.plot(df[time_resolution], df[metric_name], lw=2, color=color)
                plt.xticks([])
                ax.spines["right"].set_visible(False)
                ax.spines["top"].set_visible(False)
                plt.legend(loc="upper left", facecolor="white")

    @arthur_excepted("failed to generate drift series")
    def drift_series(
        self,
        attribute_names: Union[str, List[str]],
        drift_metric: str = "PSI",
        time_resolution: str = "day",
    ):
        """Generates a line series visualization of data drift metrics for selected attributes.

        .. code-block:: python

            # plot the KLDivergence drift of features X1, X2, and X3 by the hour
            arthur_model.viz.drift_series(["X1", "X2", "X3"], drift_metric="KLDivergence", time_resolution="hour")

        :raises ArthurUserError: failed due to user error
        :raises ArthurUnexpectedError: failed due to an unexpected error
        """
        fig = plt.figure(figsize=(20, 4))
        ax = fig.add_subplot(1, 1, 1)
        labels = []

        # if attribute_names is a string, convert to list with one element
        if isinstance(attribute_names, str):
            attribute_names = [attribute_names]

        # Define list of palettes to be used
        color_palettes = _drift_color_palettes(len(attribute_names), len(self.models))

        for model_index, model in enumerate(self.models):
            if model.is_batch:
                time_resolution = "batch_id"
            timestamp_query = _batch_timestamp_query()
            drift_query = {
                "properties": attribute_names,
                "num_bins": 20,
                "rollup": time_resolution,
                "base": {"source": "reference"},
                "target": {"source": "inference"},
                "metric": drift_metric,
            }
            if model.is_batch:
                drift_df = pd.DataFrame(model.query(drift_query, query_type="drift"))
                timestamp_df = pd.DataFrame(model.query(timestamp_query))
                df = (
                    drift_df.rename(columns={"rollup": time_resolution})
                    .join(timestamp_df.set_index(time_resolution), on=time_resolution)
                    .sort_values(by="timestamp")
                    .set_index(time_resolution)
                )
            else:
                df = (
                    pd.DataFrame(model.query(drift_query, query_type="drift"))
                    .sort_values(by="rollup")
                    .set_index("rollup")
                )

            for attribute_name in attribute_names:
                labels.append(_label(model, attribute_name))

            # base_color_palette = style.categorical_palette
            color_palette = color_palettes[model_index]
            # This darkens the color for each successive model version. There are probably better way of doing this
            # but this is just a hack to get it to work for now.
            # for color_tuple in base_color_palette:
            #     color_palette.append((color_tuple[0] / (2 ** model_index), color_tuple[1] / (2 ** model_index), color_tuple[2] / (2 ** model_index), color_tuple[3]))
            df[attribute_names].plot(ax=ax, color=color_palette, lw=8)
            plt.ylabel(drift_metric)
            plt.xticks([])
            plt.xlabel("")
            ax.spines["right"].set_visible(False)
            ax.spines["top"].set_visible(False)
        plt.legend(labels, loc="upper left", facecolor="white")

    def _timeline_categorical(self, model, attribute_name, time_resolution="day"):
        """ """
        print(_label(model, attribute_name))
        if model.is_batch:
            time_resolution = "timestamp"
            query = _timeline_categorical_batch_query(attribute_name, time_resolution)
        else:
            query = _timeline_categorical_streaming_query(
                attribute_name, time_resolution
            )
        response = model.query(query)
        df = (
            pd.DataFrame(response)
            .sort_values(by=time_resolution)
            .pivot(index=time_resolution, columns=attribute_name, values="count")
            .fillna(0.0)
        )

        sns.set(style="white")
        fig = plt.figure(figsize=(20, 4))
        ax = fig.add_subplot(1, 1, 1)
        plt.stackplot(
            df.index.values,
            df.values.T.tolist(),
            labels=df.columns.values,
            colors=style.categorical_palette,
        )
        plt.xticks([])
        plt.ylabel("Count")
        plt.legend(loc="upper right")
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        plt.show()

    def _timeline_continuous(self, model, attribute_name, time_resolution="day"):
        """ """
        print(_label(model, attribute_name))
        if model.is_batch:
            time_resolution = "timestamp"
            query = _timeline_continuous_batch_query(attribute_name, time_resolution)
        else:
            query = _timeline_continuous_streaming_query(
                attribute_name, time_resolution
            )
        response = model.query(query)

        dfs = []
        for group in response:
            temp_df = pd.DataFrame(group["distribution"])
            temp_df[time_resolution] = group[time_resolution]
            temp_df["count"] = utils.savgol_filter(temp_df["count"])
            dfs.append(temp_df)
        df = pd.concat(dfs).sort_values([time_resolution, "lower"])

        global_y_min = df["lower"].min()
        global_y_max = df["lower"].max()
        global_x_min = df["count"].min()
        global_x_max = df["count"].max()

        sns.set(style="white", rc={"axes.facecolor": (0, 0, 0, 0)})
        # Initialize the FacetGrid object
        pal = style.continuous_palette
        g = sns.FacetGrid(
            df,
            col=time_resolution,
            hue=time_resolution,
            aspect=0.2,
            height=8,
            palette=pal,
            subplot_kws={
                "xlim": (global_x_max, global_x_min),
                "ylim": (global_y_min, global_y_max),
            },
        )

        # Draw the densities in a few steps
        g.map(
            plt.plot, "count", "lower", clip_on=False, color="white", lw=4.5, alpha=0.6
        )
        g.map(plt.fill_between, "count", "lower", clip_on=False, lw=2)

        # Set the subplots to overlap
        g.fig.subplots_adjust(wspace=-0.55, left=0.125)
        g.set_titles("")
        g.set(yticks=[])
        g.set(xticks=[])
        g.set(ylabel="")
        g.set(xlabel="")
        g.despine(bottom=True, left=True)


def _metric_series_streaming_query(
    metric_name,
    time_resolution,
    ground_truth_property,
    predicted_property,
    classifier_threshold,
):
    """Generates query for fetching metrics for streaming model."""
    query = {
        "select": [
            {
                "function": metric_name,
                "alias": metric_name,
                "parameters": {
                    "ground_truth_property": ground_truth_property,
                    "predicted_property": predicted_property,
                    "threshold": classifier_threshold,
                },
            },
            {
                "function": "roundTimestamp",
                "alias": time_resolution,
                "parameters": {
                    "property": "inference_timestamp",
                    "time_interval": time_resolution,
                },
            },
        ],
        "group_by": [{"alias": time_resolution}],
        "order_by": [{"alias": time_resolution, "direction": "desc"}],
    }
    return query


def _metric_series_batch_query(
    metric_name, ground_truth_property, predicted_property, classifier_threshold
):
    """Generates query for fetching metrics for batch model."""
    query = {
        "select": [
            {
                "function": metric_name,
                "alias": metric_name,
                "parameters": {
                    "ground_truth_property": ground_truth_property,
                    "predicted_property": predicted_property,
                    "threshold": classifier_threshold,
                },
            },
            {
                "function": "max",
                "parameters": {"property": "inference_timestamp"},
                "alias": "timestamp",
            },
            {"property": "batch_id"},
        ],
        "group_by": [{"property": "batch_id"}],
        "order_by": [{"alias": "timestamp", "direction": "desc"}],
    }
    return query


def _timeline_continuous_streaming_query(attribute_name, time_resolution):
    """Generates a query for continuous attribute in a streaming model."""
    return {
        "select": [
            {
                "function": "distribution",
                "alias": "distribution",
                "parameters": {"property": attribute_name, "num_bins": 50},
            },
            {
                "function": "roundTimestamp",
                "alias": time_resolution,
                "parameters": {
                    "property": "inference_timestamp",
                    "time_interval": time_resolution,
                },
            },
        ],
        "group_by": [{"alias": time_resolution}],
        "order_by": [{"alias": time_resolution, "direction": "desc"}],
    }


def _timeline_continuous_batch_query(attribute_name, time_resolution):
    """Generates a query for continuous attribute in a batch model."""
    return {
        "select": [
            {
                "function": "distribution",
                "alias": "distribution",
                "parameters": {"property": attribute_name, "num_bins": 50},
            },
            {
                "function": "max",
                "alias": "timestamp",
                "parameters": {"property": "inference_timestamp"},
            },
            {"property": "batch_id"},
        ],
        "group_by": [{"property": "batch_id"}],
        "order_by": [{"alias": "timestamp", "direction": "desc"}],
    }


def _timeline_categorical_streaming_query(attribute_name, time_resolution):
    """Generates a query for categorical attribute in a streaming model."""
    return {
        "select": [
            {"property": attribute_name},
            {"function": "count"},
            {
                "function": "roundTimestamp",
                "alias": time_resolution,
                "parameters": {
                    "property": "inference_timestamp",
                    "time_interval": time_resolution,
                },
            },
        ],
        "group_by": [{"property": attribute_name}, {"alias": time_resolution}],
        "order_by": [
            {"property": attribute_name},
            {"alias": time_resolution, "direction": "desc"},
        ],
    }


def _timeline_categorical_batch_query(attribute_name, time_resolution):
    """Generates a query for categorical attribute in a batch model."""
    return {
        "select": [
            {"property": attribute_name},
            {"function": "count"},
            {"property": "batch_id"},
            {
                "function": "max",
                "parameters": {"property": "inference_timestamp"},
                "alias": "timestamp",
            },
        ],
        "group_by": [{"property": attribute_name}, {"property": "batch_id"}],
        "order_by": [
            {"property": attribute_name},
            {"alias": "timestamp", "direction": "desc"},
        ],
    }


def _batch_timestamp_query():
    return {
        "select": [
            {
                "function": "max",
                "alias": "timestamp",
                "parameters": {"property": "inference_timestamp"},
            },
            {"property": "batch_id"},
        ],
        "group_by": [{"property": "batch_id"}],
        "order_by": [{"alias": "timestamp", "direction": "desc"}],
    }


def _label(model, label):
    model_name = model.display_name
    if len(model_name) <= 20:
        short_model_name = model_name
    else:
        short_model_name = model_name[0:17] + "..."
    return (
        "(" + str(model.version_sequence_num) + ")__" + short_model_name + "__" + label
    )


def _alpha_color(color):
    return (color[0], color[1], color[2], style.alpha)


def _drift_color_palettes(num_attributes, num_models):
    max_attributes = len(style.multi_categorical_palette_names)
    if num_attributes > max_attributes:
        raise UserValueError(
            f"Too many attributes to display. The maximum is {max_attributes}, you wanted to see {num_attributes}"
        )

    color_palettes = []
    color_palette_names = style.multi_categorical_palette_names[0:num_attributes]
    for i in range(num_models):
        color_palettes.append([])
    for color_palette_name in color_palette_names:
        base_palette = sns.color_palette(
            palette=color_palette_name, n_colors=num_models
        )
        for i, base_palette_color in enumerate(base_palette):
            color_palettes[i].append(_alpha_color(base_palette_color))

    return color_palettes



import datetime
import uuid
from arthurai.common.constants import Stage, OutputType
from arthurai.common.exceptions import (
    ExpectedParameterNotFoundError,
    MethodNotApplicableError,
)
from arthurai.core.models import ArthurModel
import pandas as pd
from typing import Optional, Any, Tuple, Callable
import pytz


def log_prediction(arthur_model: ArthurModel) -> Callable[[Any], Any]:
    """Decorator to log the inputs and prediction of a model to Arthur.

    .. deprecated:: 3.21.1

    :param arthur_model: A previously-saved ArthurModel object

    Note, the prediction function to be decorated can take optional arguments for logging;
    these should be passed as kwargs into the function to be decorated.

    :param inference_timestamp: A timestamp in ISO 8601 format
    :param partner_inference_id: A unique identifier for an inference

    :return: Tuple of (model_prediction, inference_id)
    """
    if arthur_model.output_type not in [OutputType.Regression, OutputType.Multiclass]:
        raise MethodNotApplicableError(
            "Logging decorator supports only Regression and Multiclass output types."
        )

    def decorator_send(func):
        def wrapper_send(*args, **kwargs):
            try:
                timestamp = kwargs.pop("inference_timestamp")
            except KeyError:
                timestamp = datetime.datetime.now(pytz.utc)

            try:
                inference_id = kwargs.pop("partner_inference_id")
            except KeyError:
                inference_id = str(uuid.uuid4())

            output = func(*args)

            inference_data = _format_input_data(*args, arthur_model)
            inference_data.update(_format_predicted_value_data(output, arthur_model))

            arthur_model.send_inferences(
                [
                    {
                        "inference_data": inference_data,
                        "partner_inference_id": inference_id,
                        "inference_timestamp": timestamp,
                    }
                ]
            )

            return output, inference_id

        return wrapper_send

    return decorator_send


def _get_attr_ordering(arthur_model):
    """Return attribute names in order sorted by attr.position"""
    model_input_attrs = [
        (attr.name, attr.position)
        for attr in arthur_model.get_attributes(Stage.ModelPipelineInput)
    ]
    model_input_attrs.sort(key=lambda t: t[1])
    input_attribute_names = [name for (name, pos) in model_input_attrs]

    predicted_val_attrs = [
        (attr.name, attr.position)
        for attr in arthur_model.get_attributes(Stage.PredictedValue)
    ]
    predicted_val_attrs.sort(key=lambda t: t[1])
    predicted_val_attribute_names = [name for (name, pos) in predicted_val_attrs]
    return input_attribute_names, predicted_val_attribute_names


def _format_input_data(input_vec, arthur_model):
    """Arrange in ModelPipelineInput data into proper order and construct dict."""
    if isinstance(input_vec, pd.DataFrame):
        return input_vec.to_dict(orient="rows")[0]
    else:
        raise MethodNotApplicableError(str(input_vec.__class__) + " not yet supported.")


def _format_predicted_value_data(model_output, arthur_model):
    """Arrange PredictedValue data into proper order and construct dict."""
    try:
        # wrap in an iterable so the model output can be zipped with column names
        # primarily relevant for regression models with a single output
        model_output.__iter__()
    except AttributeError:
        model_output = [model_output]

    _validate_predicted_value_attrs(arthur_model)
    input_attribute_names, predicted_val_attribute_names = _get_attr_ordering(
        arthur_model
    )
    return dict(zip(predicted_val_attribute_names, model_output))


def _validate_predicted_value_attrs(arthur_model):
    """Check if the positions of the PredictedValue attributes have been set. If any are None
    then we can't complete the request.
    """
    attr_positions = [
        attr.position for attr in arthur_model.get_attributes(Stage.PredictedValue)
    ]
    if any([p is None for p in attr_positions]):
        raise ExpectedParameterNotFoundError(
            "Unable to send inference. All PredictedValue attributes must have `position` assigned."
        )



import logging
import os
import shutil
import tempfile
import uuid
import zipfile
from datetime import datetime
from io import BufferedRandom, BufferedReader, BytesIO
from os import PathLike
from pathlib import Path, PurePath
from sys import getsizeof
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Tuple, Union
import pandas as pd

from arthurai.common.constants import InputType, Stage
from arthurai.common.exceptions import UserValueError
from arthurai.core import util

#  imports ArthurModel for type checking, required due to circular import
if TYPE_CHECKING:
    from arthurai.core.models import ArthurModel

logger = logging.getLogger(__name__)

# fixed constants
MEGABYTE_SIZE = 1024 * 1024


class DatasetService:
    COUNTS = "counts"
    SUCCESS = "success"
    FAILURE = "failure"
    TOTAL = "total"
    FAILURES = "failures"
    DEFAULT_MAX_IMAGE_DATA_BYTES = 300000000  # 300 MB
    MAX_ROWS_PER_FILE = 500_000
    ROW_GROUP_SIZE = 5_000
    # TODO: a more robust approach would be to write a single row group, measure the file size, and use that to decide
    #  how big to make the files

    @staticmethod
    def chunk_image_set(
        directory_path: str,
        image_attribute: str,
        max_image_data_bytes: int = DEFAULT_MAX_IMAGE_DATA_BYTES,
    ) -> str:
        return DatasetService.chunk_image_set_with_directory_path_or_files(
            image_attribute,
            directory_path=directory_path,
            max_image_data_bytes=max_image_data_bytes,
        )

    @staticmethod
    def chunk_image_set_with_directory_path_or_files(
        image_attribute: str,
        directory_path: Optional[str] = None,
        files: Optional[Union[List[str], List[Path], List[PathLike]]] = None,
        max_image_data_bytes: int = DEFAULT_MAX_IMAGE_DATA_BYTES,
    ) -> str:
        """Takes in a directory path with parquet and/or json files containing image attributes.
        Divides images up into 300MB chunks, then zipped, the parquet/json file is also split up to match.
        The files will have random filename, and image zips will have matching name.

        """
        # make output dir for storing all chunks. At end will get:
        # tmp_dir/
        #    123.parquet, 123.zip, 456.parquet, 456.zip
        # TODO remove print statements, add logs to indicate processing, can take a while to run
        if directory_path is None and files is None:
            raise UserValueError(
                "Exactly 1 of the params 'directory_path' and 'files' must be provided, but neither were supplied"
            )
        if directory_path is not None and files is not None:
            raise UserValueError(
                "1 and only 1 of the params 'directory_path' and 'files' must be provided, but both were supplied"
            )

        output_dir = tempfile.mkdtemp()
        if directory_path is not None:
            files = util.retrieve_parquet_files(directory_path)
            files += util.retrieve_json_files(directory_path)
            if not files:
                raise UserValueError(
                    "The directory supplied does not contain any parquet or json files to upload"
                )

        # loop through each file
        if files is None:
            return output_dir

        for file in files:
            # keep track of where we are in the file, in case file needs to be split
            # to match image chunk
            cur_size = 0
            last_df_chunk_index = 0
            cur_img_dir = tempfile.mkdtemp(prefix=output_dir + "/")

            file_name = file
            if not (isinstance(file, str) or isinstance(file, PathLike)):
                raise UserValueError(
                    "Buffered files cannot be uploaded for CV Image models. Please provide a path to the file instead of an IO buffer."
                )
            file_suffix = ""

            if ".parquet" in str(file_name):
                file_suffix = ".parquet"
                df = pd.read_parquet(file)
            elif ".json" in str(file_name):
                file_suffix = ".json"
                df = pd.read_json(file)
            else:
                continue

            if image_attribute not in df:
                # TODO should we raise exception here instead?
                logger.warning(
                    f"Found file with missing image attribute, not including in reference set: {file_name}"
                )
                continue

            # loop through each row in file
            for cur_df_index, image_path in enumerate(df[image_attribute]):
                # verify image exists
                if not os.path.exists(image_path):
                    # TODO raise error here?
                    logger.warning(
                        f"Image does not exist for row, not including in reference set: {image_path}"
                    )
                    continue

                # move image to temp dir
                image_path = PurePath(image_path)
                temp_image_path = os.path.join(cur_img_dir, image_path.name)
                shutil.copyfile(image_path, temp_image_path)
                img_bytes = os.path.getsize(temp_image_path)
                cur_size += img_bytes

                # if we have reached max image file size, save and start new chunk
                if cur_size >= max_image_data_bytes:
                    chunk_name = str(uuid.uuid4())

                    # create chunk
                    df_chunk = df.iloc[last_df_chunk_index : cur_df_index + 1]
                    # replace image attribute with just the filename, no path
                    df_chunk[image_attribute] = df_chunk[image_attribute].apply(
                        lambda x: PurePath(x).name
                    )
                    df_chunk_filename = f"{chunk_name}{file_suffix}"
                    df_chunk_path = os.path.join(output_dir, df_chunk_filename)

                    if file_suffix == ".parquet":
                        df_chunk.to_parquet(df_chunk_path)
                    elif file_suffix == ".json":
                        df_chunk.to_json(df_chunk_path)

                    # zip images
                    image_zip_path = os.path.join(output_dir, chunk_name)
                    shutil.make_archive(image_zip_path, "zip", cur_img_dir)

                    # reset for next chunk
                    shutil.rmtree(cur_img_dir)
                    cur_img_dir = tempfile.mkdtemp(prefix=output_dir + "/")
                    cur_size = 0
                    last_df_chunk_index = cur_df_index + 1
            # we have reached end of current file, close off the current chunk before next file
            # TODO maybe pull this into function so no repeated code, but so many things to pass in
            chunk_name = str(uuid.uuid4())

            # create the final chunk
            df_chunk = df.iloc[last_df_chunk_index : cur_df_index + 1]
            # replace image attribute with just the filename, no path
            df_chunk[image_attribute] = df_chunk[image_attribute].apply(
                lambda x: PurePath(x).name
            )
            df_chunk_filename = f"{chunk_name}{file_suffix}"
            df_chunk_path = os.path.join(output_dir, df_chunk_filename)

            if file_suffix == ".parquet":
                df_chunk.to_parquet(df_chunk_path)
            elif file_suffix == ".json":
                df_chunk.to_json(df_chunk_path)

            # zip images
            image_zip_path = os.path.join(output_dir, chunk_name)
            shutil.make_archive(image_zip_path, "zip", cur_img_dir)

            # clean up
            shutil.rmtree(cur_img_dir)
        return output_dir

    @staticmethod
    def files_size(
        files: List[Union[str, Path, PathLike, BufferedReader, BufferedRandom]],
        model_input_type: InputType,
    ) -> int:
        all_files = files.copy()

        # extra image zip file should have same path and name as parquet file if model is image model
        if model_input_type == InputType.Image:
            for f in files:
                if isinstance(f, str) or isinstance(f, PathLike):
                    f_path = Path(f)
                    all_files.append(
                        Path(os.path.join(f_path.parent, f_path.stem) + ".zip")
                    )
        total_size_bytes = 0
        for file in all_files:
            try:
                if isinstance(file, str) or isinstance(file, PathLike):
                    total_size_bytes += os.path.getsize(file)
                elif isinstance(file, BufferedReader) or isinstance(
                    file, BufferedRandom
                ):
                    total_size_bytes += getsizeof(file)
            except FileNotFoundError:
                pass
        return total_size_bytes

    @staticmethod
    def send_files_from_dir_iteratively(
        model: "ArthurModel",
        directory_path: str,
        endpoint: str,
        upload_file_param_name: str,
        additional_form_params: Optional[Dict[str, Any]] = None,
        retries: int = 0,
    ) -> Tuple[List[Any], Dict[str, Any]]:
        """Sends parquet or json files iteratively from a specified directory to a specified url for a given model

        :param retries:                Number of times to retry the request if it results in a 400 or higher response code
        :param model:                  the :py:class:`!arthurai.client.apiv2.model.ArthurModel`
        :param directory_path:         local path containing parquet and/or json files to send
        :param endpoint:               POST url endpoint to send files to
        :param upload_file_param_name: key to use in body with each attached file
        :param additional_form_params: dictionary of additional form file params to send along with parquet or json file

        :raises MissingParameterError: the request failed

        :returns A list of files which failed to upload
        """
        file_types = "parquet"
        files = util.retrieve_parquet_files(directory_path)
        # don't search for json files if we're specifically uploading something like "inferences.parquet"
        if not upload_file_param_name.endswith(".parquet"):
            file_types = "json or parquet"
            files += util.retrieve_json_files(directory_path)
        if len(files) == 0:
            raise UserValueError(
                f"Could not find any {file_types} files int the given directory path: '{directory_path}'"
            )

        return DatasetService.send_files_iteratively(
            model,
            files,
            endpoint,
            upload_file_param_name,
            additional_form_params,
            retries,
        )

    @staticmethod
    def send_files_iteratively(
        model: "ArthurModel",
        files: List[Union[str, Path, PathLike, BufferedReader, BufferedRandom]],
        endpoint: str,
        upload_file_param_name: str,
        additional_form_params: Optional[Dict[str, Any]] = None,
        retries: int = 0,
    ) -> Tuple[List[Any], Dict[str, Any]]:
        if len(files) == 0:
            raise UserValueError(f"No files given to send")

        total_size = DatasetService.files_size(files, model.input_type)
        logger.info(
            f"Starting upload ({total_size / MEGABYTE_SIZE:.3f} MB in {len(files)} files), depending"
            f" on data size this may take a few minutes"
        )

        failed_files: List[Union[str, Path, PathLike]] = []
        succeeded_files: List[Union[str, Path, PathLike]] = []

        counts: Dict[str, int] = {
            DatasetService.SUCCESS: 0,
            DatasetService.FAILURE: 0,
            DatasetService.TOTAL: 0,
        }
        failures: List[Any] = []

        for file in files:
            if isinstance(file, str) or isinstance(file, PathLike):
                file_path = Path(file)
                with open(file, "rb") as open_file:
                    DatasetService._send_file(
                        model,
                        open_file,
                        endpoint,
                        upload_file_param_name,
                        additional_form_params,
                        retries,
                        counts,
                        failures,
                        succeeded_files,
                        failed_files,
                        file_path=file_path,
                    )
            else:
                DatasetService._send_file(
                    model,
                    file,
                    endpoint,
                    upload_file_param_name,
                    additional_form_params,
                    retries,
                    counts,
                    failures,
                    succeeded_files,
                    failed_files,
                )

        file_upload_info = {
            DatasetService.COUNTS: counts,
            DatasetService.FAILURES: failures,
        }

        # Only log failed or succeeded files if they exist
        if len(failed_files) > 0:
            logger.error(f"Failed to upload {len(failed_files)} files")
        if len(succeeded_files) > 0:
            logger.info(f"Successfully uploaded {len(succeeded_files)} files")
        return failed_files, file_upload_info

    @staticmethod
    def _send_file(
        model: "ArthurModel",
        open_file: Union[BytesIO, BufferedReader, BufferedRandom],
        endpoint: str,
        upload_file_param_name: str,
        additional_form_params: Optional[Dict[str, Any]],
        retries: int,
        counts: Dict[str, int],
        failures: List[Any],
        succeeded_files: List[Union[str, Path, PathLike]],
        failed_files: List[Union[str, PathLike]],
        file_path: Optional[Union[str, Path, PathLike]] = None,
    ) -> None:
        if file_path is None:
            file_path = f"{datetime.now().strftime('%Y%m%d%H%M%S')}_{uuid.uuid4()}"
            logger.debug(
                f"No file path provided, assigning temporary file name: '{file_path}'."
            )
            if model.input_type == InputType.Image:
                raise UserValueError(
                    "Buffered files cannot be uploaded for CV Image models. Please provide a path to the file instead of an IO buffer."
                )
        expected_keys = {
            DatasetService.SUCCESS,
            DatasetService.FAILURE,
            DatasetService.TOTAL,
        }
        headers = {"Content-Type": "multipart/form-data"}
        form_parts = {} if additional_form_params is None else additional_form_params
        form_parts.update({upload_file_param_name: open_file})
        # add corresponding image data if image model
        if model.input_type == InputType.Image:
            # image zip file has same path and name as parquet or json file
            image_zip_name = (
                str(os.path.join(Path(file_path).parent, Path(file_path).stem)) + ".zip"
            )
            image_zip_file = open(image_zip_name, "rb")
            form_parts.update(
                {"image_data": (image_zip_name, image_zip_file, "application/zip")}
            )
        resp = model._client.post(
            endpoint,
            json=None,
            files=form_parts,
            headers=headers,
            return_raw_response=True,
            retries=retries,
        )
        if resp.status_code == 201:
            logger.info(f"Uploaded completed: {file_path}")
            succeeded_files.append(file_path)
        elif resp.status_code == 207:
            logger.info(f"Upload completed: {file_path}")
            result: Dict[str, Dict[str, int]] = resp.json()
            # ensure the response is in the correct format
            if (
                DatasetService.COUNTS in result
                and DatasetService.FAILURES in result
                and set(result[DatasetService.COUNTS].keys()) == expected_keys
            ):
                counts[DatasetService.SUCCESS] += result[DatasetService.COUNTS][
                    DatasetService.SUCCESS
                ]
                counts[DatasetService.FAILURE] += result[DatasetService.COUNTS][
                    DatasetService.FAILURE
                ]
                counts[DatasetService.TOTAL] += result[DatasetService.COUNTS][
                    DatasetService.TOTAL
                ]
                failures.append(result[DatasetService.FAILURES])
            else:
                failures.append(result)
        else:
            logger.error(f"Failed to upload file: {resp.text}")
            failed_files.append(file_path)
            failures.append(resp.json())
            counts[DatasetService.FAILURE] += 1
            counts[DatasetService.TOTAL] += 1

        # close image zip
        if model.input_type == InputType.Image:
            image_zip_file.close()
            try:
                os.remove(image_zip_file.name)
            except Exception:
                logger.warning(
                    f"Failed to delete temporary image file at {image_zip_file.name}"
                )


class ImageZipper:
    def __init__(self):
        self.temp_file = tempfile.NamedTemporaryFile()
        self.zip = zipfile.ZipFile(self.temp_file.name, "w")

    def add_file(self, path: str):
        self.zip.write(path)

    def get_zip(self):
        self.zip.close()
        return self.temp_file

    def __del__(self):
        self.zip.close()
        self.temp_file.close()



API_PREFIX = "/api/v3"
API_PREFIX_V4 = "/api/v4"


class ListableStrEnum:
    """
    This class operates similar to Enum but passes mypy type checking.
    The members are accessed directly rather than by `.name` or `.value`.
    """

    @classmethod
    def list(cls):
        """
        Lists all attributes in alphabetical order
        """
        members = [
            getattr(cls, attr)
            for attr in dir(cls)
            if not callable(getattr(cls, attr)) and not attr.startswith("__")
        ]
        return sorted(members)


class InputType(ListableStrEnum):
    Tabular = "TABULAR"
    Image = "IMAGE"
    NLP = "NLP"
    TimeSeries = "TIME_SERIES"


class OutputType(ListableStrEnum):
    Regression = "REGRESSION"
    Multiclass = "MULTICLASS"
    Multilabel = "MULTILABEL"
    ObjectDetection = "OBJECT_DETECTION"
    TokenSequence = "TOKEN_SEQUENCE"
    RankedList = "RANKED_LIST"


class ValueType(ListableStrEnum):
    String = "STRING"
    Integer = "INTEGER"
    Float = "FLOAT"
    Image = "IMAGE"
    Boolean = "BOOLEAN"
    Timestamp = "TIMESTAMP"
    Unstructured_Text = (
        "UNSTRUCTURED_TEXT"  # don't remove old one, backward compatibility
    )
    BoundingBox = "BOUNDING_BOX"
    Tokens = "TOKENS"
    TokenLikelihoods = "TOKEN_LIKELIHOODS"
    RankedList = "RANKED_LIST"
    StringArray = "ARRAY(STRING)"
    TimeSeries = "TIME_SERIES"


class Stage(ListableStrEnum):
    ModelPipelineInput = "PIPELINE_INPUT"
    PredictFunctionInput = "PREDICT_FUNCTION_INPUT"
    PredictedValue = "PREDICTED_VALUE"
    NonInputData = "NON_INPUT_DATA"
    GroundTruth = "GROUND_TRUTH"
    GroundTruthClass = "GROUND_TRUTH_CLASS"


class TextDelimiter(ListableStrEnum):
    """This class contains patterns that can be used as text_delimiter for NLP models."""

    NOT_WORD = "\W+"
    """``"\W+"`` Splits on any character that is not a word.
    
    Ex: ``"this  is,aaaa,,,,,test!" = ["this", "is", "aaaa", "test", ""]``
    """

    WHITESPACE = "\s+"
    """``"\s+"`` Splits on whitespace.
    
    Ex: ``"this  is,a test! " = ["this", "is,a", "test!", ""]``"""

    COMMA = ","
    """``","`` Splits on a single comma.
    
    Ex: ``"this ,is,,a,test" = ["this ", "is", "", "a", "test"]``"""

    COMMA_PLUS = ",+"
    """``",+"`` Splits on one or more commas.
    
    Ex: ``"this ,is,,a,test" = ["this ", "is", "a", "test"]``"""

    PIPE = "\|"
    """``"\|"`` Splits on a single pipe.
    
    Ex: ``"this |is||a|test" = ["this ", "is", "", "a", "test"]``"""

    PIPE_PLUS = "\|+"
    """``"\|+"`` Splits on one or more pipes.
    
    Ex: ``"this |is||a|test" = ["this ", "is", "a", "test"]``"""


class Enrichment(ListableStrEnum):
    """This class contains constants for the names of enrichments"""

    AnomalyDetection = "anomaly_detection"
    BiasMitigation = "bias_mitigation"
    Explainability = "explainability"
    Hotspots = "hotspots"


class EnrichmentStatus(ListableStrEnum):
    """This class contains constants for the statuses of enrichments"""

    EnrichmentStatusDisabled = "Disabled"
    EnrichmentStatusPending = "Pending"
    EnrichmentStatusTraining = "Training"
    EnrichmentStatusReady = "Ready"
    EnrichmentStatusFailed = "Failed"


class ImageResponseType(ListableStrEnum):
    """Valid image response types supported by Arthur API"""

    RawImage = "raw_image"
    ResizedImage = "resized_image"
    Thumbnail = "thumbnail"
    LimeExplanation = "lime_explanation"


class ImageContentType(ListableStrEnum):
    """Valid image content types supported by Arthur API"""

    Png = "image/png"
    Jpeg = "image/jpeg"
    Gif = "image/gif"
    Tiff = "image/tiff"


IMAGE_FILE_EXTENSION_MAP = {
    ImageContentType.Png: ".png",
    ImageContentType.Jpeg: ".jpg",
    ImageContentType.Gif: ".gif",
    ImageContentType.Tiff: ".tff",
}


class AccuracyMetric(ListableStrEnum):
    Accuracy = "accuracy"
    Recall = "recall"
    F1 = "f1"
    Precision = "precision"


class Role(ListableStrEnum):
    User = "User"
    ModelOwner = "Model Owner"
    Administrator = "Administrator"


class ModelStatus(ListableStrEnum):
    Ready = "Ready"
    CreationFailed = "CreationFailed"
    Creating = "Creating"
    Pending = "Pending"
    Archived = "Archived"
    Archiving = "Archiving"
    ArchiveFailed = "ArchiveFailed"
    Unknown = "Unknown"


class InferenceType(ListableStrEnum):
    INFERENCE_DATA = "inference_data"
    REFERENCE_DATA = "reference_data"
    GROUND_TRUTH_DATA = "ground_truth_data"


class TimestampInferenceType(ListableStrEnum):
    INFERENCE_TIMESTAMP = "inference_timestamp"
    GROUND_TRUTH_TIMESTAMP = "ground_truth_timestamp"


ONBOARDING_SPINNER_MESSAGE = "Waiting for the model to be ready to accept inferences."
ONBOARDING_UPDATE_MESSAGE = (
    "We are still working on getting your model ready to accept inferences..."
)
DEFAULT_SERVICE_ACCOUNT = "default_sdk_service_account"

ENRICHMENTS_SPINNER_MESSAGE = (
    "Waiting for the model's enrichments to be ready to receive inferences."
)
ENRICHMENTS_UPDATE_MESSAGE = "We are still working on getting your model's enrichments ready to accept inferences..."
ENRICHMENT_READY_OR_FAILED_STATES = [
    EnrichmentStatus.EnrichmentStatusDisabled,
    EnrichmentStatus.EnrichmentStatusTraining,
    EnrichmentStatus.EnrichmentStatusReady,
    EnrichmentStatus.EnrichmentStatusFailed,
]

PARQUET_INGESTIBLE_INPUT_TYPES = [
    InputType.Tabular,
    InputType.Image,
    InputType.NLP,
]  # excludes TimeSeries
PARQUET_INGESTIBLE_OUTPUT_TYPES = [
    OutputType.Regression,
    OutputType.Multiclass,
    OutputType.Multilabel,
    OutputType.ObjectDetection,
    OutputType.TokenSequence,
]  # excludes RankedList



import json
from io import BufferedReader, BufferedRandom
from os import PathLike
from pathlib import Path
import pyarrow.parquet as pq
from pyarrow.lib import ArrowInvalid
from typing import Dict, List, Optional, Union, Any

import arthurai
import numpy as np
from arthurai.common.constants import Stage, OutputType, ValueType


def check_has_bias_attrs(model):
    """
    Returns True if and only if the model has any attributes being monitored for bias by Arthur
    """
    for attr in model.attributes:
        if attr.monitor_for_bias:
            return True
    return False


def check_attr_is_bias(model, attr_name: str):
    """
    Returns True if and only if the model has an attribute with the given name `attr_name` being monitored for bias by
    Arthur
    """
    attr = model.get_attribute(attr_name)
    return attr.monitor_for_bias


def get_positive_predicted_class(model):
    """
    Checks if model is a binary classifier. Returns False if multiclass, otherwise returns the name of the positive
    predicted attribute
    """
    if model.output_type != OutputType.Multiclass:
        return "Not a classifier model."
    predicted_value_attributes = model.get_attributes(stage=Stage.PredictedValue)
    if len(predicted_value_attributes) != 2:
        return "Not a binary classifier."
    for attr in predicted_value_attributes:
        if attr.is_positive_predicted_attribute:
            return attr.name


"""
Token Sequence Output Type Related Helpers
"""

EPSILON = 1e-3


def tensors_to_arthur_inference(
    log_tensor: np.array, id_to_vocab: Dict[int, str], num_probs: int
) -> List[Dict[str, float]]:
    """
    Convert an array of log probabilities with shape (sequence length, vocab size) to an Arthur Token Likelihoods
    formatted object.

    :param log_tensor: array of log probabilities for a generated sequence
    :param id_to_vocab: mapping from token ids to vocab strings
    :param num_probs: the number of tokens at each index in the sequence to include (up to 5)
    """
    if num_probs > 5:
        raise ValueError("num_probs must be <= 5")

    # TODO: this function should probably be batch
    if log_tensor.ndim != 2:
        raise ValueError("expected token probabilities to be two-dimensional")

    arthur_likelihoods = []
    probabilities = np.exp(log_tensor)
    invalid_probs = np.abs(np.sum(probabilities, axis=1) - 1) >= EPSILON
    if np.sum(invalid_probs) != 0:
        raise ValueError("all token probabilities must sum to 1")

    # get num_probs argmax
    top_indices = np.argpartition(probabilities, -num_probs, axis=1)[:, -num_probs:]
    top_values = probabilities[np.vstack(np.arange(len(probabilities))), top_indices]
    for v, i in zip(top_values, top_indices):
        arthur_likelihoods.append(
            {id_to_vocab[index]: value.item() for value, index in zip(v, i)}
        )
    return arthur_likelihoods


def _text_attr(name: str, stage: Stage):
    return arthurai.ArthurAttribute(
        name=name,
        stage=stage,
        position=0,
        categorical=True,
        value_type=ValueType.Unstructured_Text,
        is_unique=False if stage == Stage.PredictedValue else True,
    )


def _tokens_attr(name: str, stage: Stage, position: int = 0):
    return arthurai.ArthurAttribute(
        name=name,
        stage=stage,
        value_type=ValueType.Tokens,
        categorical=False,
        is_unique=False,
        position=position,
    )


def _token_likelihoods_attr(name: str, position: int = 0):
    return arthurai.ArthurAttribute(
        name=name,
        stage=Stage.PredictedValue,
        value_type=ValueType.TokenLikelihoods,
        categorical=False,
        is_unique=False,
        position=position,
    )


def _image_attr(name: str):
    return arthurai.ArthurAttribute(
        name=name,
        stage=Stage.ModelPipelineInput,
        value_type=ValueType.Image,
        categorical=False,
        position=0,
    )


def _get_json_data(
    file_path: Union[str, Path, PathLike, BufferedReader, BufferedRandom]
) -> Optional[Any]:
    """Gets data from json file

    Also validates file_path refers to a json file.

    :param file_path: path of json file
    :return: Python object with loaded json data, None if file_path does not refer to a json file
    """
    if isinstance(file_path, (str, Path, PathLike)):
        if not str(file_path).endswith(".json"):
            return None
        with open(file_path, "r") as json_file:
            json_data = json.loads(json_file.read())
    else:
        try:
            json_data = json.loads(file_path.read())
        except UnicodeDecodeError:
            return None
    return json_data


def _get_parquet_file(
    file_path: Union[str, Path, PathLike, BufferedReader, BufferedRandom]
) -> Optional[pq.ParquetFile]:
    """Gets ParquetFile from parquet file

    Also validates file_path refers to a parquet file.

    :param file_path: path of parquet file
    :return: ParquetFile corresponding to file_path, None if file_path does not refer to a parquet file
    """
    try:
        return pq.ParquetFile(file_path)
    except (ArrowInvalid, TypeError):
        return None



import datetime
from dateutil.parser import parse, ParserError
import logging
import re
from typing import Any

import numpy as np
import pandas as pd
from pandas import Series
from pandas.core.dtypes.common import is_float_dtype

from arthurai import ArthurAttribute, util as arthur_util
from arthurai.common.constants import ValueType
from arthurai.common.exceptions import (
    UnexpectedTypeError,
    UnexpectedValueError,
    UserTypeError,
    UserValueError,
)
from arthurai.core.util import (
    NumpyEncoder,
    is_bool_like,
    is_date_like,
    is_float_like,
    is_int_like,
    is_list_like,
    is_str_like,
    is_valid_datetime_obj,
    can_cast,
)

logger = logging.getLogger(__name__)


def validate_series_data_type(series: Series, attr: ArthurAttribute) -> None:
    """Validates if the series data type matches the attribute's Arthur value type

    Raises a UserTypeError if there is a data type mismatch where the series data type is not valid for the
    passed attribute's value type. Logs a warning if the series data type cannot be inferred because of None objects
    or empty lists or dicts.

    :param series: the Series to infer the data type from
    :param attr: the ArthurAttribute that the Series corresponds to
    :raises: UnexpectedValueError: A parameter is None.
    :raises: UserTypeError: Found data type mismatch between series and attr.
    """

    if series is None or attr is None:
        raise UnexpectedValueError("Parameter was None.")

    # standardize series
    series = series.replace([np.inf, -np.inf], np.nan)
    nans_present = series.isnull().values.any()
    series = series.dropna()
    if len(series) != 0:
        if (
            is_float_dtype(series)
            and nans_present
            and attr.value_type == ValueType.Integer
        ):
            if len(series) > 0 and np.array_equal(
                series, series.astype(int)
            ):  # int col with nans received as floats
                series = series.astype(pd.Int64Dtype())

        # validate data type mismatch
        sample_datum = series.iloc[0]
        ensure_obj_matches_attr_value_type(sample_datum, attr.value_type, attr.name)
    else:
        logger.warning(
            f"The data type of column {attr.name} cannot be validated because the column only has "
            f"None objects."
        )


def obj_value_type_mismatch_err(
    attr_name: str, expected: str, obj_type: str
) -> UserTypeError:
    """Returns UserTypeError with data type mismatch message
    :param: attr_name: Name of attribute whose value type has been mismatched with a column's data type
    :param: expected: String representing expected data type for the attribute
    :param: obj_type: String representing the object type that didn't match expected
    :raises: UnexpectedTypeError: Parameter was unexpectedly a None object.
    :returns: UserTypeError with mismatch message.
    """
    if attr_name is None or expected is None or obj_type is None:
        raise UnexpectedTypeError("Cannot pass None parameters.")
    return UserTypeError(
        f"Expected {expected} data type for attribute {attr_name} but got object of type {obj_type}. "
        f"Please correct the data type mismatch by changing either the value_type field of the "
        f"attribute or the data type of the column."
    )


def ensure_obj_matches_attr_value_type(
    obj: Any, value_type: ValueType, source: str
) -> None:
    """Validates the object data type matches the Arthur value type

    Raises a UserTypeError if the object data type does not match any valid data type corresponding to
    the Arthur Value Type of the passed attribute.

    :param obj: Object whose data type should be compared
    :param value_type: ValueType to validate against obj
    :param source: String representation of source of data for clear error messages
    :raises: UserTypeError when obj's data type does not match attr's value_type Value Type.
    """
    if source is None or obj is None or value_type is None:
        raise UnexpectedTypeError("Cannot pass None parameters.")

    if value_type == ValueType.Float:
        if not is_float_like(obj):
            raise obj_value_type_mismatch_err(source, "float", str(type(obj)))
    elif value_type == ValueType.Integer:
        if not is_int_like(obj):
            raise obj_value_type_mismatch_err(source, "int", str(type(obj)))
    elif (
        value_type == ValueType.String
        or value_type == ValueType.Image
        or value_type == ValueType.Unstructured_Text
    ):
        if not is_str_like(obj):
            raise obj_value_type_mismatch_err(source, "str", str(type(obj)))
        elif value_type == ValueType.String and is_valid_datetime_obj(
            obj
        ):  # validate str obj isn't a valid timestamp
            raise UserTypeError(
                f"Column {source} parses to valid timestamps but its attribute value_type is "
                f"{ValueType.String}, not {ValueType.Timestamp}. Please change the attribute "
                f"value_type and ensure the column consists of timezone aware Datetime objects."
            )
        elif can_cast(float, obj) or can_cast(int, obj):
            logger.warning(
                f"Column {source} has a attribute value_type of {ValueType.String} and its data are "
                f"string objects, but at least one of its data parses to a valid int or float. Consider "
                f"casting the column's data to integer or float types and changing the attribute "
                f"value_type if all the data are numbers."
            )
    elif value_type == ValueType.Boolean:
        if not is_bool_like(obj):
            raise obj_value_type_mismatch_err(source, "bool", str(type(obj)))
    elif value_type == ValueType.Timestamp:
        if not is_date_like(obj):
            raise obj_value_type_mismatch_err(
                source, "datetime or pd.Timestamp", str(type(obj))
            )
        else:
            datetime_obj = NumpyEncoder.convert_value(obj)
            arthur_util.format_timestamp(datetime_obj, col_name=source)
    elif value_type == ValueType.Tokens:  # expected type is list[str]
        list_obj = get_first_elem_if_valid_list(
            obj, source, "list[str]"
        )  # validates list
        if list_obj is not None and not is_str_like(list_obj):  # validates list element
            raise obj_value_type_mismatch_err(
                source, "list[str]", f"list[{str(type(list_obj))}]"
            )
    elif value_type == ValueType.TokenLikelihoods:
        validate_token_likelihoods_type(obj, source)
    elif (
        value_type == ValueType.BoundingBox
    ):  # expected type is list[list[Union[float, int]]]
        list_obj = get_first_elem_if_valid_list(
            obj, source, "list[list[Union[int, float]]]"
        )
        if list_obj is not None:
            if not is_list_like(list_obj):
                raise obj_value_type_mismatch_err(
                    source,
                    "list[list[Union[int, float]]]",
                    f"list[{str(type(list_obj))}]",
                )
            elif (
                len(list_obj) != 6 and not len(list_obj) == 0
            ):  # warning logged in _get_first_elem_if_valid_list if length is 0
                raise UserValueError(
                    f"Bounding box column {source} should have six values in the nested "
                    f"list but an entry in your column had {len(list_obj)} values. For a full "
                    f"description of Arthur's bounding box format see: "
                    f"https://docs.arthur.ai/docs/cv-onboarding#formatting-bounding-boxes."
                )
            else:  # validate inner list
                nested_obj = get_first_elem_if_valid_list(
                    list_obj, source, "list[list[Union[int, float]]]", "list"
                )
                if nested_obj is not None and not (
                    is_float_like(nested_obj) or is_int_like(nested_obj)
                ):
                    raise obj_value_type_mismatch_err(
                        source,
                        "list[list[Union[int, float]]]",
                        f"list[list[{str(type(nested_obj))}]]",
                    )


def get_first_elem_if_valid_list(
    obj: Any, source: str, expected: str, outer_type: str = ""
) -> Any:
    """Validates obj is list-like and non-empty and returns the first element.

    Raises a UserTypeError if obj is not list-like (is not a numpy array or python list). Logs a warning if the
    length of the list is 0 or the list has a None object.

    :param: obj: Object whose data type is being checked.
    :param: attr_name: String source of the object for clear error messages.
    :param: expected: String of expected type of the object for clear error messages.
    :param: outer_type: String representing the data type of the object obj was nested in for clear error messages. Defaults to empty string.
    :raises: UserTypeError if obj is not list-like.
    :returns: first element if obj is a list, None otherwise
    """
    if obj is None or source is None or expected is None or outer_type is None:
        raise UnexpectedValueError("Parameters cannot be None.")

    if not is_list_like(obj):  # validate list type
        if outer_type != "":
            raise obj_value_type_mismatch_err(
                source, expected, outer_type + "[" + str(type(obj)) + "]"
            )
        else:
            raise obj_value_type_mismatch_err(source, expected, str(type(obj)))
    list_obj = NumpyEncoder.convert_value(obj)
    if outer_type != "":
        outer_type = " nested"
    if len(list_obj) == 0:
        logger.warning(
            f"The length of the first{outer_type} list in {source} was 0. If you are using lists to "
            f"represent missing values instead of None objects, please change your empty lists to None "
            f"objects. The type of the list cannot be inferred from a zero-length list, so the data type "
            f"of {source} cannot be validated."
        )
        return None
    elif obj[0] is None:
        logger.warning(
            f"The first element in the first{outer_type} list in {source} was None. The type of the list cannot be "
            f"inferred from a None object, so the data type of {source} cannot be validated."
        )
        return None
    else:
        return list_obj[0]


def validate_token_likelihoods_type(obj: Any, source: str) -> None:
    """Validates obj data type matches TokenLikelihoods Value Type

    Raises a UserTypeError if the object's data type does not match any valid data type corresponding to the
    TokenLikelihoods Value Type. Logs a warning if any of the nested types—list, dict, or dict keys—are length 0
    or None objects. The valid data type for TokenLikelihoods is list[dict[str, Union[int, float]]].
    :param obj: Object to validate.
    :param source: String representation of the source of obj for clear error messages.
    :raises: UserTypeError when obj's data type does not match a TokenLikelihoods Value Type data type.
    """
    if obj is None or source is None:
        raise UnexpectedTypeError("Parameters cannot be None.")
    list_obj = get_first_elem_if_valid_list(
        obj, source, "list[dict[str, Union[int, float]]]"
    )
    if list_obj is not None:  # validate list element type is dict
        if not isinstance(list_obj, dict):
            raise obj_value_type_mismatch_err(
                source,
                "list[dict[str, Union[int, float]]]",
                f"list[{str(type(list_obj))}]",
            )
        if len(list_obj) == 0:
            logger.warning(
                f"The first dict in the first list in {source} was empty. The data type "
                f"cannot be validated for this column because no type can be inferred for the dictionary "
                f"keys or values."
            )
        else:  # validate key and value types of dict
            key, val = next(iter(list_obj.items()))
            if key is None or val is None:
                logger.warning(
                    f"A key or value in the first dict in the first list in {source} was None. The data "
                    f"type cannot be validated for this column because a type can't be inferred from a "
                    f"None object."
                )
            elif not is_str_like(key):
                raise obj_value_type_mismatch_err(
                    source,
                    "list[dict[str, Union[int, float]]]",
                    f"list[dict[{str(type(key))}, {str(type(val))}]]",
                )
            elif not (is_float_like(val) or is_int_like(val)):
                raise obj_value_type_mismatch_err(
                    source,
                    "list[dict[str, Union[int, float]]]",
                    f"list[dict[str, {str(type(val))}]]",
                )


def validate_attr_names(col_names: pd.Index) -> None:
    """Validates attribute names are correctly formatted

    Validates attribute names are made of alphanumeric characters and underscores and don't start with a number.

    :param: col_names: pd.Index of attribute names to be checked
    :raises: UserValueError: At least one attribute name is not correctly formatted
    """
    attr_name_regex = re.compile(r"^[A-Za-z_][A-Za-z0-9_]+$")
    misformatted_cols = []
    for col in col_names:
        if not attr_name_regex.match(col):
            misformatted_cols.append(col)
    if len(misformatted_cols) != 0:
        raise UserValueError(
            f"The following column names are misformatted: {str(misformatted_cols)[1:-1]}. Please "
            f"ensure each name consists of only alphanumeric characters and underscores and "
            f"doesn't start with a number."
        )


def valid_rec_obj(obj: Any) -> bool:
    """Validates recommendation object formatting

    Validates obj is a dict with an "item_id" key that has a string value
    :returns: True if obj meets criteria, False otherwise
    """
    return isinstance(obj, dict) and is_str_like(dict(obj).get("item_id"))
